from collections.abc import Sequence as SequenceCollection
import deepchem as dc
import numpy as np
try:
    import torch
    import torch.nn as nn
    import torch.nn.functional as F
except ModuleNotFoundError:
    raise ImportError('These classes require PyTorch to be installed.')
from typing import List, Union, Callable, Any
from deepchem.data import Dataset
from deepchem.feat.mol_graphs import ConvMol
from deepchem.metrics import to_one_hot
from deepchem.utils.typing import OneOrMany
import deepchem.models.torch_models.layers as torch_layers
from deepchem.utils.pytorch_utils import get_activation
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy


class TrimGraphOutput(nn.Module):
    """Trim the output to the correct number of samples.

    GraphGather always outputs fixed size batches.  This layer trims the output
    to the number of samples that were in the actual input tensors.
    """

    def __init__(self, **kwargs):
        super(TrimGraphOutput, self).__init__(**kwargs)

    def forward(self, inputs):
        n_samples = torch.squeeze(inputs[1])
        return inputs[0][0:n_samples]


class _GraphConvTorchModel(nn.Module):
    """
    Graph Convolutional Models.

    This class implements the graph convolutional model from the
    following paper [1]_. These graph convolutions start with a per-atom set of
    descriptors for each atom in a molecule, then combine and recombine these
    descriptors over convolutional layers.
    following [1]_.

    All arguments have the same meaning as in GraphConvModel.

    Example
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> from deepchem.models.torch_models import _GraphConvTorchModel
    >>> batch_size = 10
    >>> out_channels = 2
    >>> raw_smiles = ['CCC', 'C']
    >>> from rdkit import Chem
    >>> mols = [Chem.MolFromSmiles(s) for s in raw_smiles]
    >>> featurizer = dc.feat.graph_features.ConvMolFeaturizer()
    >>> mols = featurizer.featurize(mols)
    >>> multi_mol = dc.feat.mol_graphs.ConvMol.agglomerate_mols(mols)
    >>> atom_features = torch.from_numpy(multi_mol.get_atom_features().astype(np.float32))
    >>> degree_slice = torch.from_numpy(multi_mol.deg_slice)
    >>> membership = torch.from_numpy(multi_mol.membership)
    >>> deg_adjs = [torch.from_numpy(i) for i in multi_mol.get_deg_adjacency_lists()[1:]]
    >>> args = [atom_features, degree_slice, membership, torch.tensor(2)] + deg_adjs
    >>> model = _GraphConvTorchModel(out_channels, graph_conv_layers=[64, 64], number_input_features=[atom_features.shape[-1], 64],dense_layer_size=128,dropout=0.0,mode="classification",number_atom_features=75,n_classes=2,batch_normalize=False,uncertainty=False,batch_size=batch_size)
    >>> result = model(args)
    >>> len(result)
    3

    References
    ----------
    .. [1] Duvenaud, David K., et al. "Convolutional networks on graphs for learning molecular fingerprints."
        Advances in neural information processing systems. 2015. https://arxiv.org/abs/1509.09292
        """

    def __init__(self,
                 n_tasks: int,
                 number_input_features: List[int],
                 graph_conv_layers: List[int] = [64, 64],
                 dense_layer_size: int = 128,
                 dropout=0.0,
                 mode: str = "classification",
                 number_atom_features: int = 75,
                 n_classes: int = 2,
                 batch_normalize: bool = True,
                 uncertainty: bool = False,
                 batch_size: int = 100):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        number_input_features: int
            Number of input features to each of the Graph Conv Layer
        graph_conv_layers: list of int
            Width of channels for the Graph Convolution Layers
        dense_layer_size: int
            Width of channels for Atom Level Dense Layer after GraphPool
        dropout: list or float
            the dropout probablity to use for each layer.  The length of this list
            should equal len(graph_conv_layers)+1 (one value for each convolution
            layer, and one for the dense layer).  Alternatively this may be a single
            value instead of a list, in which case the same value is used for every
            layer.
        mode: str
            Either "classification" or "regression"
        number_atom_features: int
            75 is the default number of atom features created, but
            this can vary if various options are passed to the
            function atom_features in graph_features
        n_classes: int
            the number of classes to predict (only used in classification mode)
        batch_normalize: True
            if True, apply batch normalization to model
        uncertainty: bool
            if True, include extra outputs and loss terms to enable the uncertainty
            in outputs to be predicted
        """
        super(_GraphConvTorchModel, self).__init__()
        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")
        self.n_tasks: int = n_tasks
        self.n_classes: int = n_classes
        self.mode: str = mode
        self.uncertainty: bool = uncertainty

        if not isinstance(dropout, SequenceCollection):
            dropout = [dropout] * (len(graph_conv_layers) + 1)
        if len(dropout) != len(graph_conv_layers) + 1:
            raise ValueError('Wrong number of dropout probabilities provided')
        if uncertainty:
            if mode != "regression":
                raise ValueError(
                    "Uncertainty is only supported in regression mode")
            if any(d == 0.0 for d in dropout):
                raise ValueError(
                    'Dropout must be included in every layer to predict uncertainty'
                )

        self.graph_convs: nn.ModuleList = nn.ModuleList([
            torch_layers.GraphConv(layer_size,
                                   input_size,
                                   activation_fn=get_activation('relu'))
            for layer_size, input_size in zip(graph_conv_layers,
                                              number_input_features)
        ])

        self.batch_norms: nn.ModuleList = nn.ModuleList([
            nn.BatchNorm1d(num_features=64,
                           eps=1e-3,
                           momentum=0.99,
                           affine=True,
                           track_running_stats=True)
            if batch_normalize else nn.Identity()
            for _ in range(len(graph_conv_layers))
        ])
        self.batch_norms.append(
            nn.BatchNorm1d(num_features=dense_layer_size,
                           eps=1e-3,
                           momentum=0.99,
                           affine=True,
                           track_running_stats=True) if batch_normalize else nn.
            Identity())
        self.dropouts: nn.ModuleList = nn.ModuleList([
            nn.Dropout(rate) if rate > 0.0 else nn.Identity()
            for rate in dropout
        ])
        self.graph_pools: nn.ModuleList = nn.ModuleList(
            [torch_layers.GraphPool() for _ in graph_conv_layers])
        self.dense: nn.Linear = nn.Linear(64, dense_layer_size)
        self.dense_act = F.relu
        self.graph_gather = torch_layers.GraphGather(
            batch_size=batch_size, activation_fn=get_activation('tanh'))
        self.trim = TrimGraphOutput()
        if self.mode == 'classification':
            self.reshape_dense: nn.Linear = nn.Linear(dense_layer_size * 2,
                                                      n_tasks * n_classes)
        else:
            self.regression_dense: nn.Linear = nn.Linear(
                dense_layer_size * 2, n_tasks)
            if self.uncertainty:
                self.uncertainty_dense: nn.Linear = nn.Linear(
                    dense_layer_size * 2, n_tasks)
                self.uncertainty_trim = TrimGraphOutput()

    def forward(self,
                inputs: OneOrMany[torch.Tensor],
                training=False) -> List[torch.Tensor]:
        """
        Parameters
        ----------
        inputs: OneOrMany[torch.Tensor]
        Should contain tensors [atom_features, degree_slice, membership, n_samples] and deg_adjs

        Returns
        -------
        List[torch.Tensor]
            Output as per use case : regression/classification
        """
        atom_features: torch.Tensor = inputs[0]
        degree_slice: torch.Tensor = inputs[1]
        membership: torch.Tensor = inputs[2].to(torch.int64)
        n_samples: torch.Tensor = inputs[3].to(torch.int64)
        deg_adjs: List[torch.Tensor] = [
            deg_adj.to(torch.int64) for deg_adj in inputs[4:]
        ]

        in_layer: torch.Tensor = atom_features
        for i in range(len(self.graph_convs)):
            gc_in: List[torch.Tensor] = [in_layer, degree_slice, membership
                                        ] + deg_adjs
            gc1: torch.Tensor = self.graph_convs[i](gc_in)
            if self.batch_norms[i] is not None:
                gc1 = self.batch_norms[i](gc1)
            if training and self.dropouts[i] is not None:
                gc1 = self.dropouts[i](gc1)
            gp_in: List[torch.Tensor] = [gc1, degree_slice, membership
                                        ] + deg_adjs
            in_layer = self.graph_pools[i](gp_in)
        dense: torch.Tensor = self.dense(in_layer)
        denseact: torch.Tensor = self.dense_act(dense)
        if self.batch_norms[-1] is not None:
            denseact = self.batch_norms[-1](denseact)
        if training and self.dropouts[-1] is not None:
            denseact = self.dropouts[-1](denseact)
        neural_fingerprint: torch.Tensor = self.graph_gather(
            [denseact, degree_slice, membership] + deg_adjs)
        if self.mode == 'classification':
            logits: torch.Tensor = torch.reshape(
                self.reshape_dense(neural_fingerprint),
                (-1, self.n_tasks, self.n_classes))
            logits = self.trim([logits, n_samples])
            output: torch.Tensor = F.softmax(logits, dim=2)
            outputs: List[torch.Tensor] = [output, logits, neural_fingerprint]
        else:
            output = self.regression_dense(neural_fingerprint)
            output = self.trim([output, n_samples])
            if self.uncertainty:
                log_var: torch.Tensor = self.uncertainty_dense(
                    neural_fingerprint)
                log_var = self.uncertainty_trim([log_var, n_samples])
                var: torch.Tensor = torch.exp(log_var)
                outputs = [output, var, output, log_var, neural_fingerprint]
            else:
                outputs = [output, neural_fingerprint]

        return outputs


class GraphConvModel(TorchModel):
    """Graph Convolutional Models.

    This class implements the graph convolutional model from the
    following paper [1]_. These graph convolutions start with a per-atom set of
    descriptors for each atom in a molecule, then combine and recombine these
    descriptors over convolutional layers.
    following [1]_.

    Example
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> from deepchem.models.torch_models import GraphConvModel
    >>> featurizer = dc.feat.ConvMolFeaturizer()
    >>> tasks = ["outcome"]
    >>> mols = ["C", "CO", "CC"]
    >>> X = featurizer(mols)
    >>> y = np.array([0, 1, 0])
    >>> dataset = dc.data.NumpyDataset(X, y)
    >>> classification_metric = dc.metrics.Metric(dc.metrics.roc_auc_score, np.mean, mode="classification")
    >>> batch_size = 10
    >>> model = GraphConvModel(len(tasks), number_input_features=[75, 64], batch_size=batch_size, batch_normalize=False, mode='classification')
    >>> loss = model.fit(dataset, nb_epoch=20)

    References
    ----------
    .. [1] Duvenaud, David K., et al. "Convolutional networks on graphs for
        learning molecular fingerprints." Advances in neural information processing
        systems. 2015.
    """

    def __init__(self,
                 n_tasks: int,
                 number_input_features: List[int],
                 graph_conv_layers: List[int] = [64, 64],
                 dense_layer_size: int = 128,
                 dropout: float = 0.0,
                 mode: str = "classification",
                 number_atom_features: int = 75,
                 n_classes: int = 2,
                 batch_size: int = 100,
                 batch_normalize: bool = True,
                 uncertainty: bool = False,
                 **kwargs):
        """The wrapper class for graph convolutions.

        Note that since the underlying _GraphConvKerasModel class is
        specified using imperative subclassing style, this model
        cannout make predictions for arbitrary outputs.

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        number_input_features: list of int
            Number of input features to each of the Graph Conv Layer
        graph_conv_layers: list of int
            Width of channels for the Graph Convolution Layers
        dense_layer_size: int
            Width of channels for Atom Level Dense Layer after GraphPool
        dropout: list or float
            the dropout probablity to use for each layer.  The length of this list
            should equal len(graph_conv_layers)+1 (one value for each convolution
            layer, and one for the dense layer).  Alternatively this may be a single
            value instead of a list, in which case the same value is used for every
            layer.
        mode: str
            Either "classification" or "regression"
        number_atom_features: int
            75 is the default number of atom features created, but
            this can vary if various options are passed to the
            function atom_features in graph_features
        n_classes: int
            the number of classes to predict (only used in classification mode)
        batch_normalize: True
            if True, apply batch normalization to model
        uncertainty: bool
            if True, include extra outputs and loss terms to enable the uncertainty
            in outputs to be predicted
        """
        self.mode: str = mode
        self.n_tasks: int = n_tasks
        self.n_classes: int = n_classes
        self.batch_size: int = batch_size
        self.uncertainty: bool = uncertainty
        model = _GraphConvTorchModel(
            n_tasks,
            graph_conv_layers=graph_conv_layers,
            number_input_features=number_input_features,
            dense_layer_size=dense_layer_size,
            dropout=dropout,
            mode=mode,
            number_atom_features=number_atom_features,
            n_classes=n_classes,
            batch_normalize=batch_normalize,
            uncertainty=uncertainty,
            batch_size=batch_size)
        loss: Union[SoftmaxCrossEntropy, L2Loss, Callable[[Any, Any, Any], Any]]
        if mode == "classification":
            output_types = ['prediction', 'loss', 'embedding']
            loss = SoftmaxCrossEntropy()
        else:
            if self.uncertainty:
                output_types = [
                    'prediction', 'variance', 'loss', 'loss', 'embedding'
                ]

                def loss(outputs, labels, weights):
                    output, labels = dc.models.losses._make_pytorch_shapes_consistent(
                        outputs[0], labels[0])
                    losses = torch.square(output - labels) / torch.exp(
                        outputs[1]) + outputs[1]
                    w = weights[0]
                    if len(w.shape) < len(losses.shape):
                        shape = tuple(w.shape)
                        shape = tuple(-1 if x is None else x for x in shape)
                        w = torch.reshape(
                            w,
                            shape + (1,) * (len(losses.shape) - len(w.shape)))
                    return torch.mean(losses * w)
            else:
                output_types = ['prediction', 'embedding']
                loss = L2Loss()
        super(GraphConvModel, self).__init__(model,
                                             loss,
                                             output_types=output_types,
                                             batch_size=batch_size,
                                             **kwargs)

    def default_generator(self,
                          dataset: Dataset,
                          epochs: int = 1,
                          mode: str = 'fit',
                          deterministic: bool = True,
                          pad_batches: bool = True):
        """Convert a dataset into the tensors needed for learning.
        Parameters
        ----------
        dataset: `dc.data.Dataset`
            Dataset to convert
        epochs: int, optional (Default 1)
            Number of times to walk over `dataset`
        mode: str, optional (Default 'fit')
            Ignored in this implementation.
        deterministic: bool, optional (Default True)
            Whether the dataset should be walked in a deterministic fashion
        pad_batches: bool, optional (Default True)
            If true, each returned batch will have size `self.batch_size`.
        Returns
        -------
        Iterator which walks over the batches
        """
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None and self.mode == 'classification' and not (
                        mode == 'predict'):
                    y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                        -1, self.n_tasks, self.n_classes)
                multiConvMol = ConvMol.agglomerate_mols(X_b)
                n_samples = np.array(X_b.shape[0])
                inputs = [
                    multiConvMol.get_atom_features(), multiConvMol.deg_slice,
                    np.array(multiConvMol.membership), n_samples
                ]
                for i in range(1, len(multiConvMol.get_deg_adjacency_lists())):
                    inputs.append(multiConvMol.get_deg_adjacency_lists()[i])
                yield (inputs, [y_b], [w_b])



import logging
import torch
from typing import List, Optional, Union, Tuple
from deepchem.utils import chunkify, get_dtype_memsize
from deepchem.utils.dft_utils import BaseDF, DFMol, DensityFitInfo, eval_gto, eval_gradgto, eval_laplgto, LibcintWrapper, BaseHamilton, overlap, kinetic, nuclattr, int1e, elrep, BaseXC, BaseGrid, AtomCGTOBasis, ValGrad, SpinParam
from deepchem.utils.differentiation_utils import LinearOperator
from deepchem.utils.cache_utils import Cache
from deepchem.utils.dft_utils.hamilton.orbconverter import OrbitalOrthogonalizer

logger = logging.getLogger(__name__)


class HamiltonCGTO(BaseHamilton):
    """
    Hamiltonian object of contracted Gaussian type-orbital. This class
    orthogonalizes the basis by taking the weighted eigenvectors of the
    overlap matrix, i.e. the eigenvectors divided by square root of the
    eigenvalues. The advantage of doing this is making the overlap matrix
    in Roothan's equation identity and it could handle overcomplete basis.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import AtomCGTOBasis, CGTOBasis, HamiltonCGTO
    >>> cgto = CGTOBasis(angmom=0, alphas=torch.ones(1), coeffs=torch.ones(1))
    >>> atomcgto = AtomCGTOBasis(atomz=1, bases=[cgto], pos=[[0.0, 0.0, 0.0]])

    """

    def __init__(self,
                 atombases: List[AtomCGTOBasis],
                 spherical: bool = True,
                 df: Optional[DensityFitInfo] = None,
                 efield: Optional[Tuple[torch.Tensor, ...]] = None,
                 cache: Optional[Cache] = None) -> None:
        """Initialise the HamiltonCGTO

        Parameters
        ----------
        atombases: List[AtomCGTOBasis]
            Basis information of the concerned atom.
        spherical: bool
            Is the given orbital spherical.
        df: Optional[DensityFitInfo]
            Density Fitting information of the atom.
        efield: Optional[Tuple[torch.Tensor, ...]]
            Electrostatic Force field information about the environment.
        cache: Optional[Cache]
            Optional Cache parameters.

        """
        self.atombases = atombases
        self.spherical = spherical
        self.libcint_wrapper = LibcintWrapper(atombases, spherical)
        self._orthozer = OrbitalOrthogonalizer(overlap(self.libcint_wrapper))
        self.dtype = self.libcint_wrapper.dtype
        self.device = self.libcint_wrapper.device
        self._dfoptions = df
        if df is None:
            self._df: Optional[DFMol] = None
        else:
            self._df = DFMol(df,
                             wrapper=self.libcint_wrapper,
                             orthozer=self._orthozer)

        self._efield = efield
        self.is_grid_set = False
        self.is_ao_set = False
        self.is_grad_ao_set = False
        self.is_lapl_ao_set = False
        self.xc: Optional[BaseXC] = None
        self.xcfamily = 1
        self.is_built = False

        # initialize cache
        self._cache = cache if cache is not None else Cache.get_dummy()
        self._cache.add_cacheable_params(
            ["overlap", "kinetic", "nuclattr", "efield0"])
        if self._df is None:
            self._cache.add_cacheable_params(["elrep"])

    @property
    def nao(self) -> int:
        """Number of Atomic Orbitals.

        Returns
        -------
        int
            No. of Atomic Orbitals.

        """
        return self._orthozer.nao()

    @property
    def kpts(self) -> torch.Tensor:
        """The KPOINTS specifies the Bloch vectors (k points) used to
        sample the Brillouin zone. Converging this sampling is one of
        the essential tasks in many calculations concerning the
        electronic minimization.

        Bloch sphere is a geometrical representation of the pure state
        space of a two-level quantum mechanical system (qubit).

        Returns
        -------
        torch.Tensor
            List of k-points in the Hamiltonian. Shape: (nkpts, ndim)

        """
        raise TypeError(
            "Isolated molecule Hamiltonian does not have kpts property")

    @property
    def df(self) -> Optional[BaseDF]:
        """
        Returns the density fitting object (if any) attached to this
        Hamiltonian object.

        Returns
        -------
        Optional[BaseDF]
            Returns the density fitting object (if any) attached to this
            Hamiltonian object.

        """
        return self._df

    # setups
    def build(self) -> BaseHamilton:
        """
        Construct the elements needed for the Hamiltonian.
        Heavy-lifting operations should be put here.

        Returns
        -------
        BaseHamilton
            The Hamiltonian representing the total energy operator for a system of
            interacting electrons.
        """
        # get the matrices (all (nao, nao), except el_mat)
        # these matrices have already been normalized
        with self._cache.open():

            # check the signature
            self._cache.check_signature({
                "atombases": self.atombases,
                "spherical": self.spherical,
                "dfoptions": self._dfoptions,
            })

            logger.info("Calculating the overlap matrix")
            self.olp_mat = self._cache.cache(
                "overlap", lambda: overlap(self.libcint_wrapper))
            logger.info("Calculating the kinetic matrix")
            kin_mat = self._cache.cache("kinetic",
                                        lambda: kinetic(self.libcint_wrapper))
            logger.info("Calculating the nuclear attraction matrix")
            nucl_mat = self._cache.cache("nuclattr",
                                         lambda: nuclattr(self.libcint_wrapper))
            self.nucl_mat = nucl_mat
            self.kinnucl_mat = kin_mat + nucl_mat

            # electric field integral
            if self._efield is not None:
                # (ndim, nao, nao)
                fac: float = 1.0
                for i in range(len(self._efield)):
                    fac *= i + 1

                    def intor_fcn():
                        return int1e("r0" * (i + 1), self.libcint_wrapper)

                    efield_mat_f = self._cache.cache(f"efield{i}", intor_fcn)
                    efield_mat = torch.einsum("dab,d->ab", efield_mat_f,
                                              self._efield[i])
                    self.kinnucl_mat = self.kinnucl_mat + efield_mat / fac

            if self._df is None:
                logger.info("Calculating the electron repulsion matrix")
                self.el_mat = self._cache.cache(
                    "elrep", lambda: elrep(self.libcint_wrapper))  # (nao^4)
                # TODO: decide whether to precompute the 2-eris in the new basis
                # based on the memory
                self.el_mat = self._orthozer.convert4(self.el_mat)
            else:
                logger.info("Building the density fitting matrices")
                self._df.build()
            self.is_built = True

            # orthogonalize the matrices
            self.olp_mat = self._orthozer.convert2(self.olp_mat)  # (nao2, nao2)
            self.kinnucl_mat = self._orthozer.convert2(self.kinnucl_mat)
            self.nucl_mat = self._orthozer.convert2(self.nucl_mat)

            logger.info("Setting up the Hamiltonian done")

        return self

    def setup_grid(self, grid: BaseGrid, xc: Optional[BaseXC] = None) -> None:
        """
        Setup the basis (with its grad) in the spatial grid and prepare the
        gradient of atomic orbital according to the ones required by the xc.
        If xc is not given, then only setup the grid with ao (without any
        gradients of ao)

        Parameters
        ----------
        grid: BaseGrid
            Grid used to setup this Hamilton.
        xc: Optional[BaseXC] (default None)
            Exchange Corelation functional of this Hamiltonian.

        """
        # save the family and save the xc
        self.xc = xc
        if xc is None:
            self.xcfamily = 1
        else:
            self.xcfamily = xc.family

        # save the grid
        self.grid = grid
        self.rgrid = grid.get_rgrid()
        assert grid.coord_type == "cart"

        # setup the basis as a spatial function
        logger.info("Calculating the basis values in the grid")
        self.is_ao_set = True
        self.basis = eval_gto(self.libcint_wrapper,
                              self.rgrid,
                              to_transpose=True)  # (ngrid, nao)
        self.dvolume = self.grid.get_dvolume()
        self.basis_dvolume = self.basis * self.dvolume.unsqueeze(
            -1)  # (ngrid, nao)

        if self.xcfamily == 1:  # LDA
            return

        # setup the gradient of the basis
        logger.info("Calculating the basis gradient values in the grid")
        self.is_grad_ao_set = True
        # (ndim, nao, ngrid)
        self.grad_basis = eval_gradgto(self.libcint_wrapper,
                                       self.rgrid,
                                       to_transpose=True)
        if self.xcfamily == 2:  # GGA
            return

        # setup the laplacian of the basis
        self.is_lapl_ao_set = True
        logger.info("Calculating the basis laplacian values in the grid")
        self.lapl_basis = eval_laplgto(self.libcint_wrapper,
                                       self.rgrid,
                                       to_transpose=True)  # (nao, ngrid)

    # fock matrix components
    def get_nuclattr(self) -> LinearOperator:
        """LinearOperator of the nuclear Coulomb attraction.

        Nuclear Coulomb attraction is the electrostatic force binding electrons
        to a nucleus. Positively charged protons attract negatively charged
        electrons, creating stability in quantum systems. This force plays a
        fundamental role in determining the structure and behavior of atoms,
        contributing significantly to the overall potential energy in atomic
        physics.

        Returns
        -------
        LinearOperator
            LinearOperator of the nuclear Coulomb attraction. Shape: (`*BH`, nao, nao)

        """
        return LinearOperator.m(self.nucl_mat, is_hermitian=True)

    def get_kinnucl(self) -> LinearOperator:
        """
        Returns the LinearOperator of the one-electron operator (i.e. kinetic
        and nuclear attraction). Action of a LinearOperator on a function is a
        linear transformation. In the case of one-electron operators, these
        transformations are essential for solving the Schrödinger equation and
        understanding the behavior of electrons in an atomic or molecular system.

        Returns
        -------
        LinearOperator
            LinearOperator of the one-electron operator. Shape: (`*BH`, nao, nao)

        """
        return LinearOperator.m(self.kinnucl_mat, is_hermitian=True)

    def get_overlap(self) -> LinearOperator:
        """
        Returns the LinearOperator representing the overlap of the basis.
        The overlap of the basis refers to the degree to which atomic or
        molecular orbitals in a quantum mechanical system share common space.

        Returns
        -------
        LinearOperator
            LinearOperator representing the overlap of the basis.
            Shape: (`*BH`, nao, nao)

        """
        return LinearOperator.m(self.olp_mat, is_hermitian=True)

    def get_elrep(self, dm: torch.Tensor) -> LinearOperator:
        """
        Obtains the LinearOperator of the Coulomb electron repulsion operator.
        Known as the J-matrix.

        In the context of electronic structure theory, it accounts for the
        repulsive interaction between electrons in a many-electron system. The
        J-matrix elements involve the Coulombic interactions between pairs of
        electrons, influencing the total energy and behavior of the system.

        Parameters
        ----------
        dm: torch.Tensor
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        LinearOperator
            LinearOperator of the Coulomb electron repulsion operator.
            Shape: (`*BDH`, nao, nao)

        """
        if self._df is None:
            mat = torch.einsum("...ij,ijkl->...kl", dm, self.el_mat)
            mat = (mat +
                   mat.transpose(-2, -1)) * 0.5  # reduce numerical instability
            return LinearOperator.m(mat, is_hermitian=True)
        else:
            elrep = self._df.get_elrep(dm)
            return elrep

    def get_exchange(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """
        Obtains the LinearOperator of the exchange operator.
        It is -0.5 * K where K is the K matrix obtained from 2-electron integral.

        Exchange operator is a mathematical representation of the exchange
        interaction between identical particles, such as electrons. The
        exchange operator quantifies the effect of interchanging the
        positions of two particles.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            LinearOperator of the exchange operator. Shape: (`*BDH`, nao, nao)

        """
        if self._df is not None:
            raise RuntimeError(
                "Exact exchange cannot be computed with density fitting")
        elif isinstance(dm, torch.Tensor):
            # the einsum form below is to hack PyTorch's bug #57121
            # mat = -0.5 * torch.einsum("...jk,ijkl->...il", dm, self.el_mat)  # slower
            mat = -0.5 * torch.einsum("...il,ijkl->...ijk", dm,
                                      self.el_mat).sum(dim=-3)  # faster

            mat = (mat +
                   mat.transpose(-2, -1)) * 0.5  # reduce numerical instability
            return LinearOperator.m(mat, is_hermitian=True)
        else:  # dm is SpinParam
            # using the spin-scaling property of exchange energy
            return SpinParam(
                u=self.get_exchange(2 * dm.u),  # type: ignore
                d=self.get_exchange(2 * dm.d))  # type: ignore

    def get_vext(self, vext: torch.Tensor) -> LinearOperator:
        r"""
        Returns a LinearOperator of the external potential in the grid.

        .. math::
            \mathbf{V}_{ij} = \int b_i(\mathbf{r}) V(\mathbf{r}) b_j(\mathbf{r})\ d\mathbf{r}

        External potential energy that a particle experiences in a discretized
        space or grid. In quantum mechanics or computational physics, when
        solving for the behavior of particles, an external potential is often
        introduced to represent the influence of external forces.

        Parameters
        ----------
        vext: torch.Tensor
            External potential in the grid. Shape: (`*BR`, ngrid)

        Returns
        -------
        LinearOperator
            LinearOperator of the external potential in the grid. Shape: (`*BRH`, nao, nao)

        """
        if not self.is_ao_set:
            raise RuntimeError(
                "Please call `setup_grid(grid, xc)` to call this function")
        mat = torch.einsum("...r,rb,rc->...bc", vext, self.basis_dvolume,
                           self.basis)  # (*BR, nao, nao)
        mat = self._orthozer.convert2(mat)
        mat = (
            mat + mat.transpose(-2, -1)
        ) * 0.5  # ensure the symmetricity and reduce numerical instability
        return LinearOperator.m(mat, is_hermitian=True)

    def get_vxc(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """
        Returns a LinearOperator for the exchange-correlation potential

        The exchange-correlation potential combines two effects:

        1. Exchange potential: Arises from the antisymmetry of the electron
        wave function. It quantifies the tendency of electrons to avoid each
        other due to their indistinguishability.

        2. Correlation potential: Accounts for the electron-electron
        correlation effects that arise from the repulsion between electrons.

        TODO: check if what we need for Meta-GGA involving kinetics and for
        exact-exchange

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            LinearOperator for the exchange-correlation potential. Shape: (`*BDH`, nao, nao)

        """
        assert self.xc is not None, "Please call .setup_grid with the xc object"

        densinfo = SpinParam.apply_fcn(lambda dm_: self._dm2densinfo(dm_),
                                       dm)  # value: (*BD, nr)
        potinfo = self.xc.get_vxc(densinfo)  # value: (*BD, nr)
        vxc_linop = SpinParam.apply_fcn(
            lambda potinfo_: self._get_vxc_from_potinfo(potinfo_), potinfo)
        return vxc_linop

    # interface to dm
    def ao_orb2dm(self, orb: torch.Tensor,
                  orb_weight: torch.Tensor) -> torch.Tensor:
        """Convert the atomic orbital to the density matrix.

        Parameters
        ----------
        orb: torch.Tensor
            Atomic orbital. Shape: (*BO, nao, norb)
        orb_weight: torch.Tensor
            Orbital weight. Shape: (*BW, norb)

        Returns
        -------
        torch.Tensor
            Density matrix. Shape: (*BOWH, nao, nao)

        """

        orb_w = orb * orb_weight.unsqueeze(-2)  # (*BOW, nao, norb)
        return torch.matmul(orb, orb_w.transpose(-2, -1))  # (*BOW, nao, nao)

    def aodm2dens(self, dm: torch.Tensor, xyz: torch.Tensor) -> torch.Tensor:
        """Get the density value in the Cartesian coordinate.

        Parameters
        ----------
        dm: torch.Tensor
            Density matrix. Shape: (`*BD`, nao, nao)
        xyz: torch.Tensor
            Cartesian coordinate. Shape: (`*BR`, ndim)

        Returns
        -------
        torch.Tensor
            Density value in the Cartesian coordinate. Shape: (`*BRD`)

        """

        nao = dm.shape[-1]
        xyzshape = xyz.shape
        # basis: (nao, *BR)
        basis = eval_gto(self.libcint_wrapper,
                         xyz.reshape(-1, xyzshape[-1])).reshape(
                             (nao, *xyzshape[:-1]))
        basis = torch.movedim(basis, 0, -1)  # (*BR, nao)

        # torch.einsum("...ij,...i,...j->...", dm, basis, basis)
        dens = torch.matmul(dm, basis.unsqueeze(-1))  # (*BRD, nao, 1)
        dens = torch.matmul(basis.unsqueeze(-2),
                            dens).squeeze(-1).squeeze(-1)  # (*BRD)
        return dens

    # energy of the Hamiltonian
    def get_e_hcore(self, dm: torch.Tensor) -> torch.Tensor:
        """
        Get the energy from the one-electron Hamiltonian. The input is total
        density matrix.

        Parameters
        ----------
        dm: torch.Tensor
            Total Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the one-electron Hamiltonian.

        """
        return torch.einsum("...ij,...ji->...", self.kinnucl_mat, dm)

    def get_e_elrep(self, dm: torch.Tensor) -> torch.Tensor:
        """
        Get the energy from the electron repulsion. The input is total density
        matrix.

        Parameters
        ----------
        dm: torch.Tensor
            Total Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the one-electron Hamiltonian.

        """
        elrep_mat = self.get_elrep(dm).fullmatrix()
        return 0.5 * torch.einsum("...ij,...ji->...", elrep_mat, dm)

    def get_e_exchange(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Get the energy from the exact exchange.

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        torch.Tensor
            Energy from the exact exchange.

        """
        exc_mat = self.get_exchange(dm)
        ene = SpinParam.apply_fcn(
            lambda exc_mat, dm: 0.5 * torch.einsum("...ij,...ji->...",
                                                   exc_mat.fullmatrix(), dm),
            exc_mat, dm)
        enetot = SpinParam.sum(ene)
        return enetot

    def get_e_xc(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Returns the exchange-correlation energy using the xc object given in
        ``.setup_grid()``

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix. Shape: (`*BD`, nao, nao)

        Returns
        -------
        torch.Tensor
            Exchange-correlation energy.

        """
        assert self.xc is not None, "Please call .setup_grid with the xc object"

        # obtain the energy density per unit volume
        densinfo = SpinParam.apply_fcn(lambda dm_: self._dm2densinfo(dm_),
                                       dm)  # (spin) value: (*BD, nr)
        edens = self.xc.get_edensityxc(densinfo)  # (*BD, nr)

        return torch.sum(self.grid.get_dvolume() * edens, dim=-1)

    # free parameters for variational method
    def ao_orb_params2dm(
        self,
        ao_orb_params: torch.Tensor,
        ao_orb_coeffs: torch.Tensor,
        orb_weight: torch.Tensor,
        with_penalty: Optional[float] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Convert the atomic orbital free parameters (parametrized in such a way
        so it is not bounded) to the density matrix.

        Parameters
        ----------
        ao_orb_params: torch.Tensor
            The tensor that parametrized atomic orbital in an unbounded space.
        ao_orb_coeffs: torch.Tensor
            The tensor that helps ``ao_orb_params`` in describing the orbital.
            The difference with ``ao_orb_params`` is that ``ao_orb_coeffs`` is
            not differentiable and not to be optimized in variational method.
        orb_weight: torch.Tensor
            The orbital weights.
        with_penalty: float or None
            If a float, it returns a tuple of tensors where the first element is
            ``dm``, and the second element is the penalty multiplied by the
            penalty weights. The penalty is to compensate the overparameterization
            of ``ao_orb_params``, stabilizing the Hessian for gradient calculation.

        Returns
        -------
        torch.Tensor or tuple of torch.Tensor
            The density matrix from the orbital parameters and (if ``with_penalty``)
            the penalty of the overparameterization of ``ao_orb_params``.

        Notes
        -----
        * The penalty should be 0 if ``ao_orb_params`` is from ``dm2ao_orb_params``.
        * The density matrix should be recoverable when put through ``dm2ao_orb_params``
          and ``ao_orb_params2dm``.

        """
        ao_orbq, _ = torch.linalg.qr(ao_orb_params)  # (*BD, nao, norb)
        ao_orb = ao_orbq
        dm = self.ao_orb2dm(ao_orb, orb_weight)
        if with_penalty is None:
            return dm
        else:
            # QR decomposition's solution is not unique in a way that every column
            # can be multiplied by -1 and it still a solution
            # So, to remove the non-uniqueness, we will make the sign of the sum
            # positive.
            s1 = torch.sign(ao_orbq.sum(dim=-2, keepdim=True))  # (*BD, 1, norb)
            s2 = torch.sign(ao_orb_params.sum(dim=-2, keepdim=True))
            penalty = torch.mean(
                (ao_orbq * s1 - ao_orb_params * s2)**2) * with_penalty
            return dm, penalty

    def dm2ao_orb_params(
            self, dm: torch.Tensor, norb: int
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """
        Convert from the density matrix to the orbital parameters.
        The map is not one-to-one, but instead one-to-many where there
        might be more than one orbital parameters to describe the same
        density matrix. For restricted systems, only one of the ``dm``
        (``dm.u`` or ``dm.d``) is sufficient.

        Parameters
        ----------
        dm: torch.Tensor
            The density matrix.
        norb: int
            The number of orbitals for the system.

        Returns
        -------
        tuple of 2 torch.Tensor
            The atomic orbital parameters for the first returned value and the
            atomic orbital coefficients for the second value.

        Note
        ----
        This assumes that the orbital weights always decreasing in order
        """
        mdmm = dm
        w, orbq = torch.linalg.eigh(mdmm)
        # w is ordered increasingly, so we take the last parts
        orbq_params = orbq[..., -norb:]  # (nao, norb)
        return torch.flip(orbq_params, dims=(-1,))

    # misc
    def _dm2densinfo(self, dm: torch.Tensor) -> ValGrad:
        """Gets Density Fitting Info from Density Matrix.

        Parameters
        ----------
        dm: torch.Tensor
            Density Matrix

        Returns
        -------
        ValGrad

        """
        # dm: (*BD, nao, nao), Hermitian
        # family: 1 for LDA, 2 for GGA, 3 for MGGA
        # self.basis: (ngrid, nao)
        # self.grad_basis: (ndim, ngrid, nao)

        ngrid = self.basis.shape[-2]
        batchshape = dm.shape[:-2]

        # dm @ ao will be used in every case
        dmdmt = (dm + dm.transpose(-2, -1)) * 0.5  # (*BD, nao2, nao2)
        # convert it back to dm in the cgto basis
        dmdmt = self._orthozer.unconvert_dm(dmdmt)

        # prepare the densinfo components
        dens = torch.empty((*batchshape, ngrid),
                           dtype=self.dtype,
                           device=self.device)
        gdens: Optional[torch.Tensor] = None
        lapldens: Optional[torch.Tensor] = None
        kindens: Optional[torch.Tensor] = None
        if self.xcfamily == 2 or self.xcfamily == 4:  # GGA or MGGA
            gdens = torch.empty((*dm.shape[:-2], 3, ngrid),
                                dtype=self.dtype,
                                device=self.device)  # (..., ndim, ngrid)
        if self.xcfamily == 4:  # MGGA
            lapldens = torch.empty((*batchshape, ngrid),
                                   dtype=self.dtype,
                                   device=self.device)
            kindens = torch.empty((*batchshape, ngrid),
                                  dtype=self.dtype,
                                  device=self.device)

        # It is faster to split into chunks than evaluating a single big chunk
        maxnumel = 16 * 1024**2 // get_dtype_memsize(self.basis)
        for basis, ioff, iend in chunkify(self.basis, dim=0, maxnumel=maxnumel):
            # basis: (ngrid2, nao)

            dmao = torch.matmul(basis, dmdmt)  # (ngrid2, nao)
            dens[..., ioff:iend] = torch.einsum("...ri,ri->...r", dmao, basis)

            if self.xcfamily == 2 or self.xcfamily == 4:  # GGA or MGGA
                assert gdens is not None
                if not self.is_grad_ao_set:
                    msg = "Please call `setup_grid(grid, gradlevel>=1)` to calculate the density gradient"
                    raise RuntimeError(msg)

                # summing it 3 times is faster than applying the d-axis directly
                grad_basis0 = self.grad_basis[0, ioff:iend, :]  # (ngrid2, nao)
                grad_basis1 = self.grad_basis[1, ioff:iend, :]
                grad_basis2 = self.grad_basis[2, ioff:iend, :]

                gdens[..., 0, ioff:iend] = torch.einsum("...ri,ri->...r", dmao,
                                                        grad_basis0) * 2
                gdens[..., 1, ioff:iend] = torch.einsum("...ri,ri->...r", dmao,
                                                        grad_basis1) * 2
                gdens[..., 2, ioff:iend] = torch.einsum("...ri,ri->...r", dmao,
                                                        grad_basis2) * 2

            if self.xcfamily == 4:
                assert lapldens is not None
                assert kindens is not None
                # calculate the laplacian of the density and kinetic energy density at the grid
                if not self.is_lapl_ao_set:
                    msg = "Please call `setup_grid(grid, gradlevel>=2)` to calculate the density gradient"
                    raise RuntimeError(msg)

                lapl_basis_cat = self.lapl_basis[ioff:iend, :]
                lapl_basis = torch.einsum("...ri,ri->...r", dmao,
                                          lapl_basis_cat)
                grad_grad = torch.einsum("...ri,ri->...r",
                                         torch.matmul(grad_basis0, dmdmt),
                                         grad_basis0)
                grad_grad += torch.einsum("...ri,ri->...r",
                                          torch.matmul(grad_basis1, dmdmt),
                                          grad_basis1)
                grad_grad += torch.einsum("...ri,ri->...r",
                                          torch.matmul(grad_basis2, dmdmt),
                                          grad_basis2)
                # pytorch's "...ij,ir,jr->...r" is really slow for large matrix
                # grad_grad = torch.einsum("...ij,ir,jr->...r", dmdmt, self.grad_basis[0], self.grad_basis[0])
                # grad_grad += torch.einsum("...ij,ir,jr->...r", dmdmt, self.grad_basis[1], self.grad_basis[1])
                # grad_grad += torch.einsum("...ij,ir,jr->...r", dmdmt, self.grad_basis[2], self.grad_basis[2])
                lapldens[..., ioff:iend] = (lapl_basis + grad_grad) * 2
                kindens[..., ioff:iend] = grad_grad * 0.5

        # dens: (*BD, ngrid)
        # gdens: (*BD, ndim, ngrid)
        res = ValGrad(value=dens, grad=gdens, lapl=lapldens, kin=kindens)
        return res

    def _get_vxc_from_potinfo(self, potinfo: ValGrad) -> LinearOperator:
        """obtain the vxc operator from the potential information

        Parameters
        ----------
        potinfo: ValGrad
            potential information as ValGrad.
            potinfo.value (*BD, nr)
            potinfo.grad (*BD, ndim, nr)
            potinfo.lapl (*BD, nr)

        Returns
        -------
        LinearOperator
            vxc LinearOperator"""

        # prepare the fock matrix component from vxc
        nao = self.basis.shape[-1]
        mat = torch.zeros((*potinfo.value.shape[:-1], nao, nao),
                          dtype=self.dtype,
                          device=self.device)

        # Split the r-dimension into several parts, it is usually faster than
        # evaluating all at once
        maxnumel = 16 * 1024**2 // get_dtype_memsize(self.basis)
        for basis, ioff, iend in chunkify(self.basis, dim=0, maxnumel=maxnumel):
            # basis: (nr, nao)
            vb = potinfo.value[..., ioff:iend].unsqueeze(
                -1) * basis  # (*BD, nr, nao)
            if self.xcfamily in [2, 4]:  # GGA or MGGA
                assert potinfo.grad is not None  # (..., ndim, nr)
                vgrad = potinfo.grad[..., ioff:iend] * 2
                grad_basis0 = self.grad_basis[0, ioff:iend, :]  # (nr, nao)
                grad_basis1 = self.grad_basis[1, ioff:iend, :]
                grad_basis2 = self.grad_basis[2, ioff:iend, :]
                vb += torch.einsum("...r,ra->...ra", vgrad[..., 0, :],
                                   grad_basis0)
                vb += torch.einsum("...r,ra->...ra", vgrad[..., 1, :],
                                   grad_basis1)
                vb += torch.einsum("...r,ra->...ra", vgrad[..., 2, :],
                                   grad_basis2)
            if self.xcfamily == 4:  # MGGA
                assert potinfo.lapl is not None  # (..., nrgrid)
                assert potinfo.kin is not None
                lapl = potinfo.lapl[..., ioff:iend]
                kin = potinfo.kin[..., ioff:iend]
                vb += 2 * lapl.unsqueeze(-1) * self.lapl_basis[ioff:iend, :]

            # calculating the matrix from multiplication with the basis
            mat += torch.matmul(
                self.basis_dvolume[ioff:iend, :].transpose(-2, -1), vb)

            if self.xcfamily == 4:  # MGGA
                assert potinfo.lapl is not None  # (..., nrgrid)
                assert potinfo.kin is not None
                lapl_kin_dvol = (2 * lapl + 0.5 * kin) * self.dvolume[...,
                                                                      ioff:iend]
                mat += torch.einsum("...r,rb,rc->...bc", lapl_kin_dvol,
                                    grad_basis0, grad_basis0)
                mat += torch.einsum("...r,rb,rc->...bc", lapl_kin_dvol,
                                    grad_basis1, grad_basis1)
                mat += torch.einsum("...r,rb,rc->...bc", lapl_kin_dvol,
                                    grad_basis2, grad_basis2)

        # construct the Hermitian linear operator
        mat = self._orthozer.convert2(mat)
        mat = (mat + mat.transpose(-2, -1)) * 0.5
        vxc_linop = LinearOperator.m(mat, is_hermitian=True)
        return vxc_linop

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Return the paramnames

        Parameters
        ----------
        methodname: str
            Name of the method.
        prefix: str (default "")
            Prefix of the paramnames.

        Returns
        -------
        List[str]
            Paramnames.

        """
        if methodname == "get_kinnucl":
            return [prefix + "kinnucl_mat"]
        elif methodname == "get_nuclattr":
            return [prefix + "nucl_mat"]
        elif methodname == "get_overlap":
            return [prefix + "olp_mat"]
        elif methodname == "get_elrep":
            if self._df is None:
                return [prefix + "el_mat"]
            else:
                return self._df.getparamnames("get_elrep",
                                              prefix=prefix + "_df.")
        elif methodname == "get_exchange":
            return [prefix + "el_mat"]
        elif methodname == "ao_orb2dm":
            return []
        elif methodname == "ao_orb_params2dm":
            return self.getparamnames("ao_orb2dm", prefix=prefix)
        elif methodname == "get_e_hcore":
            return [prefix + "kinnucl_mat"]
        elif methodname == "get_e_elrep":
            return self.getparamnames("get_elrep", prefix=prefix)
        elif methodname == "get_e_exchange":
            return self.getparamnames("get_exchange", prefix=prefix)
        elif methodname == "get_e_xc":
            assert self.xc is not None
            return self.getparamnames("_dm2densinfo", prefix=prefix) + \
                self.xc.getparamnames("get_edensityxc", prefix=prefix + "xc.") + \
                self.grid.getparamnames("get_dvolume", prefix=prefix + "grid.")
        elif methodname == "get_vext":
            return [prefix + "basis_dvolume", prefix + "basis"] + \
                self._orthozer.getparamnames("convert2", prefix=prefix + "_orthozer.")
        elif methodname == "get_grad_vext":
            return [prefix + "basis_dvolume", prefix + "grad_basis"]
        elif methodname == "get_lapl_kin_vext":
            return [
                prefix + "dvolume", prefix + "basis", prefix + "grad_basis",
                prefix + "lapl_basis"
            ]
        elif methodname == "get_vxc":
            assert self.xc is not None
            return self.getparamnames("_dm2densinfo", prefix=prefix) + \
                self.getparamnames("_get_vxc_from_potinfo", prefix=prefix) + \
                self.xc.getparamnames("get_vxc", prefix=prefix + "xc.")
        elif methodname == "_dm2densinfo":
            params = [prefix + "basis"] + \
                self._orthozer.getparamnames("unconvert_dm", prefix=prefix + "_orthozer.")
            if self.xcfamily == 2 or self.xcfamily == 4:
                params += [prefix + "grad_basis"]
            if self.xcfamily == 4:
                params += [prefix + "lapl_basis"]
            return params
        elif methodname == "_get_vxc_from_potinfo":
            params = [prefix + "basis", prefix + "basis_dvolume"] + \
                self._orthozer.getparamnames("convert2", prefix=prefix + "_orthozer.")
            if self.xcfamily in [2, 4]:
                params += [prefix + "grad_basis"]
            if self.xcfamily == 4:
                params += [prefix + "lapl_basis", prefix + "dvolume"]
            return params
        else:
            raise KeyError("getparamnames has no %s method" % methodname)



from transformers import BertForMaskedLM, BertForSequenceClassification, BertTokenizer, BertConfig
import torch.nn as nn
from deepchem.models.torch_models.hf_models import HuggingFaceModel
from typing import Union


class ProtBERT(HuggingFaceModel):
    """
    ProtBERT model[1].

    ProtBERT model is based on BERT architecture and the current implementation
    supports only MLM pretraining and classification mode, as described by the
    authors in HuggingFace[2]. For classfication we currently only support
    Logistic regression and a simple Feed forward neural network.

    The model converts the input protein sequence into a vector through a trained BERT tokenizer, which is then
    processed by the corresponding model based on the task. `BertForMaskedLM` is used to facilitate the MLM
    pretraining task. For the sequence classification task, we follow `BertForSequenceClassification` but change
    the classifier to either a logistic regression (LogReg) or a feed-forward neural network (FFN), depending on
    the specified `cls_name`. The FFN is a simple 2-layer network with 512 as the hidden dimension.


    Examples
    --------
    >>> import os
    >>> import tempfile
    >>> tempdir = tempfile.mkdtemp()

    >>> # preparing dataset
    >>> import pandas as pd
    >>> import deepchem as dc
    >>> protein = ["MPCTTYLPLLLLLFLLPPPSVQSKV","SSGLFWMELLTQFVLTWPLVVIAFL"]
    >>> labels = [0,1]
    >>> df = pd.DataFrame(list(zip(protein, labels)), columns=["protein", "task1"])
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(["task1"], feature_field="protein", featurizer=dc.feat.DummyFeaturizer())
    ...     dataset = loader.create_dataset(tmpfile.name)

    >>> # pretraining
    >>> from deepchem.models.torch_models.prot_bert import ProtBERT
    >>> pretrain_model_dir = os.path.join(tempdir, 'pretrain-model')
    >>> model_path = 'Rostlab/prot_bert'
    >>> pretrain_model = ProtBERT(task='mlm', HG_model_path=model_path, n_tasks=1, model_dir=pretrain_model_dir)  # mlm pretraining
    >>> pretraining_loss = pretrain_model.fit(dataset, nb_epoch=1)
    >>> del pretrain_model

    >>> finetune_model_dir = os.path.join(tempdir, 'finetune-model')
    >>> finetune_model = ProtBERT(task='classification', HG_model_path=model_path, n_tasks=1, model_dir=finetune_model_dir)
    >>> finetune_model.load_from_pretrained(pretrain_model_dir)
    >>> finetuning_loss = finetune_model.fit(dataset, nb_epoch=1)

    >>> # prediction and evaluation
    >>> result = finetune_model.predict(dataset)
    >>> eval_results = finetune_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.accuracy_score))

    References
    ----------
    .. [1] Elnaggar, Ahmed, et al. "Prottrans: Toward understanding the language of life through self-supervised learning." IEEE transactions on pattern analysis and machine intelligence 44.10 (2021): 7112-7127.
    .. [2] https://huggingface.co/Rostlab/prot_bert

    """

    def __init__(self,
                 task: str,
                 model_path: str = 'Rostlab/prot_bert',
                 n_tasks: int = 1,
                 cls_name: str = "LogReg",
                 classifier_net: Union[nn.Module, None] = None,
                 n_classes: int = 2,
                 **kwargs) -> None:
        """
        Parameters
        ----------
        task: str
            The task defines the type of learning task in the model. The supported tasks are
            - `mlm` - masked language modeling commonly used in pretraining
            - `classification` - use it for classification tasks
        model_path: str
            Path to the HuggingFace model
            - 'Rostlab/prot_bert' - Pretrained on Uniref100 dataset
            - `Rostlab/prot_bert_bfd` - Pretrained on BFD dataset
        n_tasks: int
            Number of prediction targets for a multitask learning model
        cls_name: str
            The classifier head to use for classification mode. Currently supports "FFN" and "LogReg" and custom classfier head.
        classifier_net: nn.Module, optional
            A custom classifier head to use for classification mode. The network must have input size of 1024.
        n_classes: int
            Number of classes for classification.
        """
        self.n_tasks: int = n_tasks
        tokenizer: BertTokenizer = BertTokenizer.from_pretrained(
            model_path, do_lower_case=False)
        protbert_config: BertConfig = BertConfig.from_pretrained(
            pretrained_model_name_or_path=model_path,
            vocab_size=tokenizer.vocab_size)
        model: Union[BertForMaskedLM, BertForSequenceClassification]
        if task == "mlm":
            model = BertForMaskedLM.from_pretrained(model_path)
        elif task == "classification":
            cls_head: Union[nn.Linear, nn.Sequential, nn.Module, None]
            protbert_config.num_labels = n_classes
            if n_tasks == 1:
                protbert_config.problem_type = 'single_label_classification'
            else:
                protbert_config.problem_type = 'multi_label_classification'

            if (cls_name == "LogReg"):
                cls_head = nn.Linear(1024, n_classes)
            elif (cls_name == "FFN"):
                cls_head = nn.Sequential(nn.Linear(1024, 512), nn.ReLU(),
                                         nn.Linear(512, n_classes))
            elif (cls_name == "custom"):
                cls_head = classifier_net

            else:
                raise ValueError('Invalid classifier: {}.'.format(cls_name))

            model = BertForSequenceClassification.from_pretrained(
                model_path, config=protbert_config)
            model.classifier = cls_head

        else:
            raise ValueError('Invalid task specification')
        super().__init__(model=model, task=task, tokenizer=tokenizer, **kwargs)



import warnings
import torch
try:
    import pylibxc
except (ImportError, ModuleNotFoundError) as e:
    warnings.warn(f"Failed to import pylibxc. Might not be able to use xc. {e}")
from typing import List, Tuple, Union, Optional
from deepchem.utils.dft_utils import BaseXC, ValGrad, SpinParam
from deepchem.utils.dft_utils.xc.libxc_wrapper import CalcLDALibXCPol, CalcLDALibXCUnpol, \
    CalcGGALibXCPol, CalcGGALibXCUnpol, CalcMGGALibXCUnpol, CalcMGGALibXCPol


ERRMSG = "This function cannot do broadcasting. " \
         "Please make sure the inputs have the same shape."
N_VRHO = 2  # number of xc energy derivative w.r.t. density (i.e. 2: u, d)
N_VSIGMA = 3  # number of energy derivative w.r.t. contracted gradient (i.e. 3: uu, ud, dd)


class LibXCLDA(BaseXC):
    """Local Density Approximation (LDA) wrapper for libxc.
    Local-density approximations (LDA) are a class of approximations
    to the exchange–correlation (XC) energy functional in density
    functional theory (DFT) that depend solely upon the value of the
    electronic density at each point in space.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> import torch
    >>> # create a LDA wrapper for libxc
    >>> lda = LibXCLDA("lda_x")
    >>> # create a density information
    >>> densinfo = ValGrad(value=torch.rand(2, 3, 4), grad=torch.rand(2, 3, 4, 3))
    >>> # get the exchange-correlation potential
    >>> potinfo = lda.get_vxc(densinfo)
    >>> potinfo.value.shape
    torch.Size([2, 3, 4])
    >>> edens = lda.get_edensityxc(densinfo)
    >>> edens.shape
    torch.Size([2, 3, 4])

    Attributes
    ----------
    _family: int (default 1)
        Family of the exchange-correlation functional
    _unpolfcn_wrapper: torch.autograd.Function (default CalcLDALibXCUnpol)
        Wrapper for the unpolarized LDA functional
    _polfcn_wrapper: torch.autograd.Function (default CalcLDALibXCPol)
        Wrapper for the polarized LDA functional

    """

    def __init__(self, name: str) -> None:
        """Initialize the LDA wrapper for libxc.

        Parameters
        ----------
        name: str
            Name of the exchange-correlation functional

        """
        self.libxc_unpol = pylibxc.LibXCFunctional(name, "unpolarized")
        self.libxc_pol = pylibxc.LibXCFunctional(name, "polarized")
        self._family: int = 1
        self._unpolfcn_wrapper = CalcLDALibXCUnpol  # type: ignore
        self._polfcn_wrapper = CalcLDALibXCPol  # type: ignore

    @property
    def family(self) -> int:
        """Get the family of the exchange-correlation functional.

        Returns
        -------
        int
            Family of the exchange-correlation functional

        """
        return self._family

    def get_vxc(
        self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
    ) -> Union[ValGrad, SpinParam[ValGrad]]:
        """Get the exchange-correlation potential from libxc.

        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            density information
            densinfo.value: (*BD, nr)
            densinfo.grad: (*BD, nr, ndim)

        Returns
        -------
        Union[ValGrad, SpinParam[ValGrad]]
            Potential information
            potentialinfo.value: (*BD, nr)
            potentialinfo.grad: (*BD, nr, ndim)

        """
        libxc_inps = _prepare_libxc_input(densinfo, xcfamily=self.family)
        flatten_inps = tuple(inp.reshape(-1) for inp in libxc_inps)

        # polarized case
        if not isinstance(densinfo, ValGrad):
            # outs are (vrho,) for LDA, (vrho, vsigma) for GGA each with shape
            # (nspin, *shape)
            outs = self._calc_pol(flatten_inps, densinfo.u.value.shape, 1)

        # unpolarized case
        else:
            # outs are (vrho,) for LDA, (vrho, vsigma) for GGA each with shape
            # (*shape)
            outs = self._calc_unpol(flatten_inps, densinfo.value.shape, 1)

        potinfo = _postproc_libxc_voutput(densinfo, *outs)
        return potinfo

    def get_edensityxc(self, densinfo: Union[ValGrad, SpinParam[ValGrad]]) -> \
            torch.Tensor:
        """Get the exchange-correlation energy density from libxc.

        Parameters
        ----------
        densinfo: Union[ValGrad, SpinParam[ValGrad]]
            Density information
            densinfo.value: (*BD, nr)
            densinfo.grad: (*BD, nr, ndim)

        Returns
        -------
        torch.Tensor
            Exchange-correlation energy density
            edens: (*BD, nr)

        """
        libxc_inps = _prepare_libxc_input(densinfo, xcfamily=self.family)
        flatten_inps = tuple(inp.reshape(-1) for inp in libxc_inps)

        # polarized case
        if not isinstance(densinfo, ValGrad):
            rho_u = densinfo.u.value

            edens = self._calc_pol(flatten_inps, densinfo.u.value.shape,
                                   0)[0]  # (*BD, nr)
            edens = edens.reshape(rho_u.shape)
            return edens

        # unpolarized case
        else:
            edens = self._calc_unpol(flatten_inps, densinfo.value.shape,
                                     0)[0]  # (*BD, nr)
            return edens

    def _calc_pol(self, flatten_inps: Tuple[torch.Tensor, ...], shape: torch.Size, deriv: int) ->\
            Tuple[torch.Tensor, ...]:
        """Calculate the polarized exchange-correlation potential from libxc.

        Parameters
        ----------
        flatten_inps: Tuple[torch.Tensor, ...]
            Flattened inputs for libxc
        shape: torch.Size
            Shape of the density
        deriv: int
            Derivative order

        Returns
        -------
        Tuple[torch.Tensor, ...]
            Outputs from libxc

        """

        outs = self._polfcn_wrapper.apply(*flatten_inps, deriv, self.libxc_pol)

        # tuple of (nderiv, *shape) where nderiv is the number of spin-dependent
        # values from libxc
        return tuple(out.reshape(-1, *shape) for out in outs)

    def _calc_unpol(self, flatten_inps: Tuple[torch.Tensor, ...], shape: torch.Size, deriv: int) ->\
            Tuple[torch.Tensor, ...]:
        """Calculate the unpolarized exchange-correlation potential from libxc.

        Parameters
        ----------
        flatten_inps: Tuple[torch.Tensor, ...]
            Flattened inputs for libxc
        shape: torch.Size
            Shape of the density
        deriv: int
            Derivative order

        Returns
        -------
        Tuple[torch.Tensor, ...]
            Outputs from libxc

        """

        outs = self._unpolfcn_wrapper.apply(*flatten_inps, deriv,
                                            self.libxc_unpol)

        # tuple of (*shape) where shape
        return tuple(out.reshape(shape) for out in outs)

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        return []


class LibXCGGA(LibXCLDA):
    """Generalized Gradient Approximation (GGA) wrapper for libxc.

    GGA can correct the overestimated binding energy of LDA in
    molecules and solids and extend the processing system to the
    energy and structure of the hydrogen bond system. The
    approximation greatly improves the calculation results of the
    energy related to electrons and exchange.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> import torch
    >>> # create a GGA wrapper for libxc
    >>> gga = LibXCGGA("gga_c_pbe")
    >>> # create a density information
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> densinfo = ValGrad(value=rho_u, grad=grad_u)
    >>> # get the exchange-correlation potential
    >>> potinfo = gga.get_vxc(densinfo)
    >>> potinfo.value.shape
    torch.Size([2])

    Attributes
    ----------
    _family: int (default 2)
        Family of the exchange-correlation functional
    _unpolfcn_wrapper: torch.autograd.Function (default CalcGGALibXCUnpol)
        Wrapper for the unpolarized GGA functional
    _polfcn_wrapper: torch.autograd.Function (default CalcGGALibXCPol)
        Wrapper for the polarized GGA functional

    """

    def __init__(self, name: str) -> None:
        """Initialize the LDA wrapper for libxc.

        Parameters
        ----------
        name: str
            Name of the exchange-correlation functional

        """
        self.libxc_unpol = pylibxc.LibXCFunctional(name, "unpolarized")
        self.libxc_pol = pylibxc.LibXCFunctional(name, "polarized")
        self._family: int = 2
        self._unpolfcn_wrapper = CalcGGALibXCUnpol  # type: ignore
        self._polfcn_wrapper = CalcGGALibXCPol  # type: ignore


class LibXCMGGA(LibXCLDA):
    """Meta-Generalized Gradient Approximation (MGGA) wrapper for libxc.

    Meta-GGAs typically improve upon the accuracy of GGAs as they
    additionally take into account the local kinetic energy density.
    This allows meta-GGAs to more accurately treat different chemical
    bonds (eg. covalent, metallic, and weak) compared to LDAs and GGAs.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> import torch
    >>> # create a MGGA wrapper for libxc
    >>> mgga = LibXCMGGA("mgga_x_scan")
    >>> # create a density information
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> densinfo = ValGrad(value=rho_u, grad=grad_u, lapl=lapl_u, kin=kin_u)
    >>> # get the exchange-correlation potential
    >>> potinfo = mgga.get_vxc(densinfo)
    >>> potinfo.value.shape
    torch.Size([2])

    Attributes
    ----------
    _family: int (default 4)
        Family of the exchange-correlation functional
    _unpolfcn_wrapper: torch.autograd.Function (default CalcMGGALibXCUnpol)
        Wrapper for the unpolarized MGGA functional
    _polfcn_wrapper: torch.autograd.Function (default CalcMGGALibXCPol)
        Wrapper for the polarized MGGA functional

    """

    def __init__(self, name: str) -> None:
        """Initialize the LDA wrapper for libxc.

        Parameters
        ----------
        name: str
            Name of the exchange-correlation functional

        """
        self.libxc_unpol = pylibxc.LibXCFunctional(name, "unpolarized")
        self.libxc_pol = pylibxc.LibXCFunctional(name, "polarized")
        self._family: int = 4
        self._unpolfcn_wrapper = CalcMGGALibXCUnpol  # type: ignore
        self._polfcn_wrapper = CalcMGGALibXCPol  # type: ignore


def _prepare_libxc_input(densinfo: Union[SpinParam[ValGrad], ValGrad],
                         xcfamily: int) -> Tuple[torch.Tensor, ...]:
    """Prepare the input for libxc.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import ValGrad, SpinParam
    >>> import torch
    >>> # create a density information
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> rho_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_d = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> lapl_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> kin_d = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> densinfo = SpinParam(u=ValGrad(value=rho_u, grad=grad_u, lapl=lapl_u, kin=kin_u), d=ValGrad(value=rho_d, grad=grad_d, lapl=lapl_d, kin=kin_d))
    >>> # prepare the input for libxc
    >>> inputs = _prepare_libxc_input(densinfo, 4)
    >>> len(inputs)
    9

    Parameters
    ----------
    densinfo: Union[SpinParam[ValGrad], ValGrad]
        Density information
        densinfo.value: (*BD, nr)
        densinfo.grad: (*BD, nr, ndim)
    xcfamily: int
        Family of the exchange-correlation functional

    Returns
    -------
    Tuple[torch.Tensor, ...]
        Inputs for libxc

    """
    # convert the densinfo into tuple of tensors for libxc inputs
    # the elements in the tuple is arranged according to libxc manual

    sigma_einsum = "...dr,...dr->...r"
    # polarized case
    if isinstance(densinfo, SpinParam):
        rho_u = densinfo.u.value  # (*nrho)
        rho_d = densinfo.d.value

        if xcfamily == 1:  # LDA
            return (rho_u, rho_d)

        assert densinfo.u.grad is not None
        assert densinfo.d.grad is not None
        grad_u = densinfo.u.grad  # (*nrho, ndim)
        grad_d = densinfo.d.grad

        # calculate the contracted gradient
        sigma_uu = torch.einsum(sigma_einsum, grad_u, grad_u)
        sigma_ud = torch.einsum(sigma_einsum, grad_u, grad_d)
        sigma_dd = torch.einsum(sigma_einsum, grad_d, grad_d)

        if xcfamily == 2:  # GGA
            return (rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd)

        assert densinfo.u.lapl is not None
        assert densinfo.d.lapl is not None
        assert densinfo.u.kin is not None
        assert densinfo.d.kin is not None
        lapl_u = densinfo.u.lapl
        lapl_d = densinfo.d.lapl
        kin_u = densinfo.u.kin
        kin_d = densinfo.d.kin

        if xcfamily == 4:  # MGGA
            return (rho_u, rho_d, sigma_uu, sigma_ud, sigma_dd, lapl_u, lapl_d,
                    kin_u, kin_d)

    # unpolarized case
    else:
        rho = densinfo.value

        if xcfamily == 1:
            return (rho,)

        assert densinfo.grad is not None
        grad = densinfo.grad  # (*nrho, ndim)

        # contracted gradient
        sigma = torch.einsum(sigma_einsum, grad, grad)

        if xcfamily == 2:  # GGA
            return (rho, sigma)

        assert densinfo.lapl is not None
        assert densinfo.kin is not None
        lapl = densinfo.lapl
        kin = densinfo.kin

        if xcfamily == 4:  # MGGA
            return (rho, sigma, lapl, kin)

    raise RuntimeError(f"xcfamily {xcfamily} is not implemented")


def _postproc_libxc_voutput(
        densinfo: Union[SpinParam[ValGrad], ValGrad],
        vrho: torch.Tensor,
        vsigma: Optional[torch.Tensor] = None,
        vlapl: Optional[torch.Tensor] = None,
        vkin: Optional[torch.Tensor] = None
) -> Union[SpinParam[ValGrad], ValGrad]:
    """Postprocess the output from libxc into potential information.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> import torch
    >>> # create a density information
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> lapl_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> kin_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> densinfo = ValGrad(value=rho_u, grad=grad_u, lapl=lapl_u, kin=kin_u)
    >>> # postprocess the output from libxc
    >>> potinfo = _postproc_libxc_voutput(densinfo, torch.rand((n,), dtype=torch.float64).requires_grad_())
    >>> potinfo.value.shape
    torch.Size([2])

    Parameters
    ----------
    densinfo: Union[SpinParam[ValGrad], ValGrad]
        Density information
        densinfo.value: (..., nr)
        densinfo.grad: (..., ndim, nr)
    vrho: torch.Tensor
        Density potential
    vsigma: Optional[torch.Tensor]
        Gradient potential
    vlapl: Optional[torch.Tensor]
        Laplacian potential
    vkin: Optional[torch.Tensor]
        Kinetic potential

    Returns
    -------
    Union[SpinParam[ValGrad], ValGrad]
        Potential information

    """
    # polarized case
    if isinstance(densinfo, SpinParam):
        # vrho: (2, *BD, nr)
        # vsigma: (3, *BD, nr)
        # vlapl: (2, *BD, nr)
        # vkin: (2, *BD, nr)
        vrho_u = vrho[0]
        vrho_d = vrho[1]

        # calculate the gradient potential
        vgrad_u: Optional[torch.Tensor] = None
        vgrad_d: Optional[torch.Tensor] = None
        if vsigma is not None:
            # calculate the grad_vxc
            vgrad_u = 2 * vsigma[0].unsqueeze(-2) * densinfo.u.grad + \
                vsigma[1].unsqueeze(-2) * densinfo.d.grad  # (..., 3)
            vgrad_d = 2 * vsigma[2].unsqueeze(-2) * densinfo.d.grad + \
                vsigma[1].unsqueeze(-2) * densinfo.u.grad

        vlapl_u: Optional[torch.Tensor] = None
        vlapl_d: Optional[torch.Tensor] = None
        if vlapl is not None:
            vlapl_u = vlapl[0]
            vlapl_d = vlapl[1]

        vkin_u: Optional[torch.Tensor] = None
        vkin_d: Optional[torch.Tensor] = None
        if vkin is not None:
            vkin_u = vkin[0]
            vkin_d = vkin[1]

        potinfo_u = ValGrad(value=vrho_u,
                            grad=vgrad_u,
                            lapl=vlapl_u,
                            kin=vkin_u)
        potinfo_d = ValGrad(value=vrho_d,
                            grad=vgrad_d,
                            lapl=vlapl_d,
                            kin=vkin_d)
        return SpinParam(u=potinfo_u, d=potinfo_d)

    # unpolarized case
    else:
        # all are: (*BD, nr)

        # calculate the gradient potential
        if vsigma is not None:
            vsigma = 2 * vsigma.unsqueeze(-2) * densinfo.grad  # (*BD, ndim, nr)

        potinfo = ValGrad(value=vrho, grad=vsigma, lapl=vlapl, kin=vkin)
        return potinfo



def test_uniquifier():
    from deepchem.utils import Uniquifier
    a = 1
    b = 2
    c = 3
    d = 1
    u = Uniquifier([a, b, c, a, d])
    assert u.get_unique_objs() == [1, 2, 3]


def test_get_option():
    from deepchem.utils import get_option
    options = {"a": 1, "b": 2}
    assert get_option("name", "a", options) == 1



import numpy as np
from collections import defaultdict
from deepchem.feat import Featurizer
from typing import List, Dict, Tuple, Any, Optional

try:
    import dgl
    import torch
except ImportError:
    pass


class _Realigner(object):
    """
    A private class for realigning sequencing reads to a reference sequence.
    This class provides methods for left-aligning indels, processing pileup
    information, generating pileups and reads, selecting candidate regions,
    fetching reads, building De Bruijn graphs, pruning graphs, and generating
    candidate haplotypes.

    """

    def left_align_indel(self, seq: str, pos: int,
                         indel: str) -> Tuple[int, str]:
        """
        Left align an indel by shifting it to the left as
        much as possible. This function shifts an indel to the left
        if the preceding bases match the end of the indel sequence,
        making the indel left-aligned.

        Parameters
        ----------

        seq : str
            Reference sequence as a string.
        pos : int
            Position of the indel.
        indel : str
            Indel sequence (e.g., "+2AT" or "-2").

        Returns
        -------

        Tuple[int, str]
            New position and indel string.

        """
        if indel.startswith('+'):
            indel_seq = indel[2:]
            while pos > 0 and seq[pos - 1] == indel_seq[-1]:
                pos -= 1
                indel_seq = seq[pos] + indel_seq[:-1]
            return pos, f"+{len(indel_seq)}{indel_seq}"
        elif indel.startswith('-'):
            del_len = int(indel[1:])
            while pos > 0 and seq[pos - 1] == seq[pos + del_len - 1]:
                pos -= 1
            return pos, f"-{del_len}"
        return pos, indel

    def process_pileup(
            self, pileup_info: List[Dict[str,
                                         Any]], reference_seq_dict: Dict[str,
                                                                         str],
            allele_counts: Dict[Tuple[str, int], Dict[str, Any]]) -> None:
        """
        Process pileup information to extract allele counts. This function
        processes each pileup column to count the occurrences of each
        allele at each position, updating the allele counts dictionary.

        Parameters
        ----------

        pileup_info : List[Dict[str, Any]]
            List of dictionaries containing pileup information.
        reference_seq_dict : Dict[str, str]
            Dictionary with chromosome names as keys and reference
            sequences as values.
        allele_counts : Dict[Tuple[str, int], Dict[str, Any]]
            Dictionary to store allele counts.

        """
        for pileupcolumn in pileup_info:
            ref_base = reference_seq_dict[pileupcolumn['name']][
                pileupcolumn['pos']]
            allele_count: Dict[str, Any] = {
                "reference_base": ref_base,
                "read_alleles": defaultdict(int),
                "coverage": 0
            }
            for read_seq in pileupcolumn['reads']:
                if read_seq[2]:
                    base = '-'
                    print(allele_count)
                    allele_count["read_alleles"][base] += 1
                    allele_count["coverage"] += 1
                elif read_seq[3]:
                    base = 'N'
                    allele_count["read_alleles"][base] += 1
                    allele_count["coverage"] += 1
                else:
                    indel = ""
                    if read_seq[4] > 0:
                        indel_seq = read_seq[0][read_seq[2]:read_seq[2] +
                                                read_seq[4]]
                        indel = f"+{read_seq[4]}{indel_seq}"
                        new_pos, indel = self.left_align_indel(
                            reference_seq_dict[pileupcolumn['name']]
                            [0:pileupcolumn['pos'] + read_seq[4] + 1],
                            pileupcolumn['pos'], indel)
                    elif read_seq[4] < 0:
                        indel = f"{read_seq[4]}"
                        new_pos, indel = self.left_align_indel(
                            reference_seq_dict[pileupcolumn['name']]
                            [0:pileupcolumn['pos'] + read_seq[4] + 1],
                            pileupcolumn['pos'], indel)

                    base = read_seq[0][read_seq[1]]

                    if base in {'A', 'C', 'G', 'T'}:
                        if indel:
                            allele_count["read_alleles"][base + indel] += 1
                            allele_count["coverage"] += 1
                        else:
                            allele_count["read_alleles"][base] += 1
                            allele_count["coverage"] += 1
            pos = pileupcolumn['pos'] + 1
            allele_counts[(pileupcolumn['name'], pos)] = allele_count

    def generate_pileup_and_reads(
        self, bamfiles, reference_seq_dict
    ) -> Tuple[Dict[Tuple[str, int], Dict[str, Any]], List[Any]]:
        """
        Generate pileup and reads from BAM and reference FASTA files. This
        function generates pileup information and reads from the provided
        BAM files and reference sequences, returning both allele counts and reads.

        Parameters
        ----------

        bamfile_path : str
            Path to the BAM file.
        reference_path : str
            Path to the reference FASTA file.

        Returns
        -------

        Tuple[Dict[Tuple[str, int], Dict[str, Any]], List[Any]]
            Dictionary of allele counts and list of reads.

        """
        allele_counts: Dict[Tuple[str, int], Dict[str, Any]] = {}
        reads = []

        for x in bamfiles:
            chrom = x[3]  # Reference name

            pileup_info = x[7] if len(x) > 7 else None
            for pileupcolumn in pileup_info:
                for read_seq in pileupcolumn['reads']:
                    reads.append(read_seq)

            if chrom not in reference_seq_dict:
                continue

            if pileup_info:
                self.process_pileup(pileup_info, reference_seq_dict,
                                    allele_counts)

        return allele_counts, reads

    def update_counts(self, count: int, start: int, end: int,
                      window_counts: Dict[int, int]) -> None:
        """
        Update counts in a window. This function increments the count
        for each position in the specified range by the given count.

        Parameters
        ----------

        count : int
            Count to add.
        start : int
            Start position of the window.
        end : int
            End position of the window.
        window_counts : Dict[int, int]
            Dictionary to store window counts.

        """
        for pos in range(start, end):
            window_counts[pos] += count

    def select_candidate_regions(
        self, allele_counts: Dict[Tuple[str, int], Dict[str, Any]]
    ) -> List[Tuple[str, int, int, int]]:
        """
        Select candidate regions based on allele counts. This function
        identifies candidate regions with significant allele counts
        and groups adjacent positions into regions.

        Parameters
        ----------

        allele_counts : Dict[Tuple[str, int], Dict[str, Any]]
            Dictionary of allele counts.

        Returns
        -------

        List[Tuple[str, int, int, int]]
            List of candidate regions.

        """
        window_counts: Dict[int, int] = defaultdict(int)
        chrom_positions = defaultdict(list)

        for (chrom, pos), allele_count in allele_counts.items():
            for allele, count in allele_count["read_alleles"].items():
                if allele != allele_count["reference_base"]:
                    allele_type = ""
                    if allele.startswith("+") or allele.startswith("-"):
                        allele_length = int(allele[1:])
                        allele_type = "INSERTION" if allele.startswith(
                            "+") else "DELETION"
                    elif allele == "N":
                        allele_type = "SOFT_CLIP"
                        allele_length = 1
                    else:
                        allele_type = "SUBSTITUTION"
                        allele_length = 1

                    start, end = 0, 0

                    if allele_type == "SUBSTITUTION":
                        start = pos
                        end = pos + 1
                    elif allele_type in ["SOFT_CLIP", "INSERTION"]:
                        start = pos + 1 - (allele_length - 1)
                        end = pos + allele_length
                    elif allele_type == "DELETION":
                        start = pos + 1
                        end = pos + allele_length

                    self.update_counts(count, start, end, window_counts)
                    chrom_positions[chrom].append((start, end))

        candidate_regions = []
        for chrom, positions in chrom_positions.items():
            positions.sort()
            if positions:
                start, end = positions[0]
                current_count = sum(
                    window_counts[pos] for pos in range(start, end))

                for (next_start, next_end) in positions[1:]:
                    if next_start <= end + 1:
                        end = max(end, next_end)
                        current_count += sum(
                            window_counts[pos]
                            for pos in range(next_start, next_end))
                    else:
                        candidate_regions.append(
                            (chrom, start, end, current_count))
                        start, end = next_start, next_end
                        current_count = sum(
                            window_counts[pos] for pos in range(start, end))

                candidate_regions.append((chrom, start, end, current_count))

        return candidate_regions

    def fetchreads(self, bamfiles: List[Tuple[str, Any, int, str, int, Any, int,
                                              Any]], chrom: str, start: int,
                   end: int) -> List[str]:
        """
        Fetch reads from BAM files for a specific chromosome and region.
        This function extracts reads from BAM files that overlap with the
        specified chromosome and region.

        Parameters
        ----------

        bamfiles : List[Tuple[str, Any, int, str, int, Any, int, Any]]
            List of BAM file records, where each record is a tuple containing
            information about reads in the BAM file.
        chrom : str
            Chromosome name to fetch reads from.
        start : int
            Start position of the region.
        end : int
            End position of the region.

        Returns
        -------

        List[str]
            List of read sequences that fall within the specified region.
        """
        reads: List[str] = []
        for bamfile in bamfiles:
            refname = bamfile[3]
            refstart = bamfile[4]
            refend = refstart + bamfile[2]

            if refname == chrom and refstart < end and refend > start:
                reads.append(bamfile[1])
        return reads

    def build_debruijn_graph(
        self, ref: str, reads: List[str], k: int
    ) -> Tuple[Optional[Any], Optional[Dict[str, int]], Optional[Dict[int,
                                                                      str]]]:
        """
        Build a De Bruijn graph from reference and reads. This function
        constructs a De Bruijn graph from k-mers found in the reference
        and reads, where nodes represent k-mers and edges represent
        overlaps between k-mers.

        Parameters
        ----------

        ref : str
            Reference sequence as a string.
        reads : List[str]
            List of read sequences.
        k : int
            Length of k-mers.

        Returns
        -------

        Tuple[Optional[Any], Optional[Dict[str, int]], Optional[Dict[int, str]]]
            A tuple containing:
            - De Bruijn graph (dgl.DGLGraph)
            - Dictionary mapping k-mer strings to node IDs (Dict[str, int])
            - Dictionary mapping node IDs to k-mer strings (Dict[int, str])
        """
        kmer_counts: Dict[str, int] = defaultdict(int)

        def get_kmers(sequence: str, k: int):
            for i in range(len(sequence) - k + 1):
                yield sequence[i:i + k]

        # Count k-mers in reference
        for kmer in get_kmers(ref, k):
            kmer_counts[kmer] += 1

        # Count k-mers in reads
        for read in reads:
            for kmer in get_kmers(read, k):
                kmer_counts[kmer] += 1

        kmer_to_id: Dict[str, int] = {}
        id_to_kmer: Dict[int, str] = {}
        node_id: int = 0
        edges: List[Tuple[int, int]] = []
        edge_weights: List[int] = []

        for kmer in kmer_counts:
            if kmer not in kmer_to_id:
                kmer_to_id[kmer] = node_id
                id_to_kmer[node_id] = kmer
                node_id += 1

        # Add edges between overlapping k-mers
        for kmer in kmer_counts:
            suffix = kmer[1:]
            for next_kmer in kmer_counts:
                if suffix == next_kmer[:-1]:
                    edges.append((kmer_to_id[kmer], kmer_to_id[next_kmer]))
                    edge_weights.append(kmer_counts[next_kmer])

        if not edges:
            return None, None, None

        src, dst = zip(*edges)
        G = dgl.graph((src, dst), num_nodes=node_id)
        G.ndata['weight'] = torch.tensor([kmer_counts[k] for k in kmer_counts],
                                         dtype=torch.float32)
        G.edata['weight'] = torch.tensor(edge_weights, dtype=torch.float32)

        return G, kmer_to_id, id_to_kmer

    def prune_debruijn_graph(self, G: Any, min_edge_weight: float) -> Any:
        """
        This function removes edges with weights below the specified threshold
        and removes nodes that become isolated after pruning.

        Parameters
        ----------

        G : Any
            The original De Bruijn graph.
        min_edge_weight : float
            The minimum edge weight threshold for keeping an edge.

        Returns
        -------

        Any
            The pruned subgraph.
        """
        edge_weights = G.edata['weight']
        mask = edge_weights >= min_edge_weight
        edges_to_keep = mask.nonzero(as_tuple=False).squeeze().tolist()

        # Create a subgraph with the edges to keep
        subgraph = dgl.edge_subgraph(G, edges_to_keep, relabel_nodes=False)

        # Identify and remove isolated nodes
        isolated_nodes = (subgraph.in_degrees() == 0) & (subgraph.out_degrees()
                                                         == 0)
        isolated_nodes = isolated_nodes.nonzero(
            as_tuple=False).squeeze().tolist()
        subgraph = dgl.remove_nodes(subgraph, isolated_nodes)

        return subgraph

    def candidate_haplotypes(self, G: Any, k: int,
                             id_to_kmer: Optional[Dict[int, str]]) -> List[str]:
        """
        Generate candidate haplotypes from the De Bruijn graph. This function
        traverses the De Bruijn graph to generate potential haplotypes by
        combining k-mers along paths from start to end nodes.

        Parameters
        ----------

        G : dgl.DGLGraph
            The De Bruijn graph.
        k : int
            The k-mer length.
        id_to_kmer : Dict[int, str]
            A dictionary mapping node IDs to k-mers.

        Returns
        -------

        List[str]
            Sorted list of candidate haplotypes.
        """
        haplotypes: List[str] = []
        nodes = list(G.nodes().numpy())

        if not nodes:
            return haplotypes

        if id_to_kmer is None:
            return haplotypes

        start_node = nodes[0]
        end_node = nodes[-1]

        def dfs(node: int, path: List[int]) -> None:
            path.append(node)
            if node == end_node:
                haplotype = id_to_kmer[path[0]]
                for p in path[1:]:
                    haplotype += id_to_kmer[p][-1]
                haplotypes.append(haplotype)
            else:
                for succ in G.successors(node).numpy():
                    dfs(succ, path[:])

        dfs(start_node, [])

        return sorted(haplotypes)

    def process_candidate_windows(
        self, candidate_regions: List[Tuple[str, int, int, int]],
        bamfiles: List[Any], reference_seq_dict: Dict[str, str]
    ) -> List[Tuple[str, int, int, int, int, List[str]]]:
        """
        Process candidate regions to generate candidate windows with haplotypes.

        Parameters
        ----------

        candidate_regions : List[Tuple[str, int, int, int]]
            List of candidate regions.
        bamfiles : List[Any]
            List of BAM file data.
        reference_seq_dict : Dict[str, str]
            Dictionary with chromosome names as keys and reference sequences as values.

        Returns
        -------

        List[Tuple[str, int, int, int, int, List[str]]]
            List of candidate windows with haplotypes.
        """
        candidate_windows: List[Tuple[str, int, int, int, int, List[str]]] = []

        for chrom, start, end, count in candidate_regions:
            window_reads = self.fetchreads(bamfiles, chrom, start, end)
            ref_sequence = reference_seq_dict[chrom][start:end + 1]

            kmin, kmax, step_k = 15, 21, 2
            found_graph = False

            for k in range(kmin, kmax + 1, step_k):
                dbg, kmer_to_id, id_to_kmer = self.build_debruijn_graph(
                    ref_sequence, window_reads, k)

                if dbg is None:
                    continue

                if kmer_to_id is None:
                    continue

                dbg = self.prune_debruijn_graph(dbg, min_edge_weight=2)

                if dbg.number_of_nodes() == 0:
                    continue

                found_graph = True
                candidate_haplotypes_list = self.candidate_haplotypes(
                    dbg, k, id_to_kmer)

                if candidate_haplotypes_list and candidate_haplotypes_list != [
                        ref_sequence
                ]:
                    candidate_windows.append((chrom, start, end, count, k,
                                              candidate_haplotypes_list))
                    break

            if not found_graph:
                candidate_windows.append(
                    (chrom, start, end, count, k, [ref_sequence]))

        return candidate_windows


class RealignerFeaturizer(Featurizer):
    """
    Realigns reads and generates candidate regions for variant calling.

    More features can be added to this class in the future.

    Note
    ----
    This class requires Pysam, DGL and Pytorch to be installed. Pysam can be
    used with Linux or MacOS X. To use Pysam on Windows, use Windows
    Subsystem for Linux(WSL).

    """

    def __init__(self):
        self.realigner = _Realigner()

    def decode_one_hot(self,
                       one_hot_vector: List[np.ndarray],
                       charset: List[str] = ["A", "C", "T", "G", "N"]) -> str:
        """
        Decode a one-hot encoded sequence into a string of
        nucleotides.

        Parameters
        ----------

        one_hot_vector : List[np.ndarray]
            List of one-hot encoded vectors.
        charset : Optional[List[str]]
            List of characters corresponding to the encoding.
            Default is ["A", "C", "T", "G", "N"].

        Returns
        -------

        str
            Decoded sequence as a string.

        """
        decoded_seq = []
        for vector in one_hot_vector:
            idx = np.argmax(vector)
            decoded_seq.append(charset[idx])
        return ''.join(decoded_seq)

    def _featurize(self, datapoint):
        """
        Featurizes a datapoint by generating candidate regions and reads.

        Args:
            datapoint (Tuple[str, str]): A tuple containing two strings
            representing allele counts and reads.

        Returns:
            Tuple[List[Tuple[str, int, int, int]], List[Any]]: A tuple
            containing the candidate regions and reads.

        """
        bamfile_path = datapoint[0]
        reference_file_path = datapoint[1]

        from deepchem.data import BAMLoader, FASTALoader

        bam_loader = BAMLoader(get_pileup=True)
        bam_dataset = bam_loader.create_dataset(bamfile_path)
        bamfiles = bam_dataset.X

        fasta_loader = FASTALoader(None, False, False)
        fasta_dataset = fasta_loader.create_dataset(reference_file_path)

        one_hot_encoded_sequences = fasta_dataset.X
        decoded_sequences: List[str] = []

        # Convert the one-hot encoded sequences to strings
        for seq in one_hot_encoded_sequences:
            decoded_seq = self.decode_one_hot(seq)
            decoded_sequences.append(decoded_seq)

        # Map the sequences to chrom names
        chrom_names = ["chr1", "chr2"]

        reference_seq_dict: Dict[str, str] = {
            chrom_names[i]: seq for i, seq in enumerate(decoded_sequences)
        }

        allele_counts, reads = self.realigner.generate_pileup_and_reads(
            bamfiles, reference_seq_dict)
        candidate_regions = self.realigner.select_candidate_regions(
            allele_counts)
        candidate_windows = self.realigner.process_candidate_windows(
            candidate_regions, bamfiles, reference_seq_dict)
        return candidate_windows



from abc import abstractmethod
from typing import Tuple
import numpy as np
from deepchem.utils.dft_utils.hamilton.intor.utils import CSYMM, np2ctypes, int2ctypes


class BaseSymmetry(object):
    """ Base class for symmetry operations.

    This class defines the interface for symmetry operations.

    Examples
    --------
    >>> class SemNew(BaseSymmetry):
    ...     def get_reduced_shape(self, orig_shape: Tuple[int, ...]) -> Tuple[int, ...]:
    ...         return orig_shape
    ...     @property
    ...     def code(self) -> str:
    ...         return "sn"
    ...     def reconstruct_array(self, arr: np.ndarray, orig_shape: Tuple[int, ...]) -> np.ndarray:
    ...         return arr
    >>> sym = SemNew()
    >>> sym.get_reduced_shape((2, 3, 4))
    (2, 3, 4)
    >>> sym.code
    'sn'

    """

    @abstractmethod
    def get_reduced_shape(self, orig_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        """
        Get reduced shape from original shape.

        Parameters
        ----------
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        Tuple[int, ...]
            Reduced shape of the array.
        """
        pass

    @property
    @abstractmethod
    def code(self) -> str:
        """
        Short code for this symmetry.

        Returns
        -------
        str
            Code representing the symmetry.
        """
        pass

    @abstractmethod
    def reconstruct_array(self, arr: np.ndarray,
                          orig_shape: Tuple[int, ...]) -> np.ndarray:
        """
        Reconstruct full array from reduced symmetrized array.

        Parameters
        ----------
        arr : np.ndarray
            Reduced symmetrized array.
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        np.ndarray
            Reconstructed full array.
        """
        pass


class S1Symmetry(BaseSymmetry):
    """ Class for S1 symmetry.

    This class defines the operations for S1 symmetry. (no symmetry)

    Examples
    --------
    >>> sym = S1Symmetry()
    >>> sym.get_reduced_shape((2, 3, 4))
    (2, 3, 4)
    >>> sym.code
    's1'
    >>> sym.reconstruct_array(np.array([1, 2, 3]), (3,))
    array([1, 2, 3])

    """

    def get_reduced_shape(self, orig_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        """
        Get reduced shape from original shape (no symmetry).

        Parameters
        ----------
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        Tuple[int, ...]
            Reduced shape (same as original).
        """
        return orig_shape

    @property
    def code(self) -> str:
        """
        Short code for S1 symmetry.

        Returns
        -------
        str
            Code "s1".
        """
        return "s1"

    def reconstruct_array(self, arr: np.ndarray,
                          orig_shape: Tuple[int, ...]) -> np.ndarray:
        """
        Reconstruct full array from reduced symmetrized array (no symmetry).

        Parameters
        ----------
        arr : np.ndarray
            Reduced symmetrized array.
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        np.ndarray
            Reconstructed full array (same as reduced array).
        """
        return arr


class S4Symmetry(BaseSymmetry):
    """ Class for S4 symmetry.

    This class defines the operations for S4 symmetry. (ijkl) == (jikl) == (ijlk) == (jilk)

    Examples
    --------
    >>> sym = S4Symmetry()
    >>> sym.get_reduced_shape((3, 3, 4, 4))
    (6, 10)
    >>> sym.code
    's4'
    >>> sym.reconstruct_array(np.random.rand(2, 3, 4, 4), (3, 3, 4, 4)).shape
    (3, 3, 4, 4)

    """

    def get_reduced_shape(self, orig_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        """
        Get reduced shape from original shape for S4 symmetry.
        Returned shape would be (..., i(j+1)/2, k(l+1)/2)

        Parameters
        ----------
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        Tuple[int, ...]
            Reduced shape for S4 symmetry.
        """
        self.__check_orig_shape(orig_shape)

        batchshape = orig_shape[:-4]
        ijshape = orig_shape[-4] * (orig_shape[-3] + 1) // 2
        klshape = orig_shape[-2] * (orig_shape[-1] + 1) // 2
        return (*batchshape, ijshape, klshape)

    @property
    def code(self) -> str:
        """
        Short code for S4 symmetry.

        Returns
        -------
        str
            Code "s4".
        """
        return "s4"

    def reconstruct_array(self, arr: np.ndarray,
                          orig_shape: Tuple[int, ...]) -> np.ndarray:
        """
        Reconstruct full array from reduced symmetrized array for S4 symmetry. (..., ij/2, kl/2)

        Parameters
        ----------
        arr : np.ndarray
            Reduced symmetrized array.
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Returns
        -------
        np.ndarray
            Reconstructed full array for S4 symmetry.
        """
        self.__check_orig_shape(orig_shape)

        out = np.zeros(orig_shape, dtype=arr.dtype)
        fcn = CSYMM().fills4
        fcn(np2ctypes(out), np2ctypes(arr), int2ctypes(orig_shape[-4]),
            int2ctypes(orig_shape[-2]))
        return out

    def __check_orig_shape(self, orig_shape: Tuple[int, ...]):
        """
        Check if original shape satisfies requirements for S4 symmetry.

        Parameters
        ----------
        orig_shape : Tuple[int, ...]
            Original shape of the array.

        Raises
        ------
        AssertionError
            If original shape does not satisfy requirements for S4 symmetry.
        """
        assert len(orig_shape) >= 4
        assert orig_shape[-4] == orig_shape[-3]
        assert orig_shape[-2] == orig_shape[-1]



"""
Featurizes proposed binding pockets.
"""
import numpy as np
import logging
from typing import Dict, List

from deepchem.feat import Featurizer
from deepchem.utils.coordinate_box_utils import CoordinateBox
from deepchem.utils.rdkit_utils import load_molecule

logger = logging.getLogger(__name__)


def boxes_to_atoms(
        coords: np.ndarray,
        boxes: List[CoordinateBox]) -> Dict[CoordinateBox, List[int]]:
    """Maps each box to a list of atoms in that box.

    Given the coordinates of a macromolecule, and a collection of boxes,
    returns a dictionary which maps boxes to the atom indices of the
    atoms in them.

    Parameters
    ----------
    coords: np.ndarray
        A numpy array of shape `(N, 3)`
    boxes: list
        List of `CoordinateBox` objects.

    Returns
    -------
    Dict[CoordinateBox, List[int]]
        A dictionary mapping `CoordinateBox` objects to lists of atom indices.
    """
    mapping = {}
    for box_ind, box in enumerate(boxes):
        box_atoms = []
        for atom_ind in range(len(coords)):
            atom = coords[atom_ind]
            if atom in box:
                box_atoms.append(atom_ind)
        mapping[box] = box_atoms
    return mapping


class BindingPocketFeaturizer(Featurizer):
    """Featurizes binding pockets with information about chemical
    environments.

    In many applications, it's desirable to look at binding pockets on
    macromolecules which may be good targets for potential ligands or
    other molecules to interact with. A `BindingPocketFeaturizer`
    expects to be given a macromolecule, and a list of pockets to
    featurize on that macromolecule. These pockets should be of the form
    produced by a `dc.dock.BindingPocketFinder`, that is as a list of
    `dc.utils.CoordinateBox` objects.

    The base featurization in this class's featurization is currently
    very simple and counts the number of residues of each type present
    in the pocket. It's likely that you'll want to overwrite this
    implementation for more sophisticated downstream usecases. Note that
    this class's implementation will only work for proteins and not for
    other macromolecules

    Note
    ----
    This class requires mdtraj to be installed.
    """

    residues = [
        "ALA", "ARG", "ASN", "ASP", "CYS", "GLN", "GLU", "GLY", "HIS", "ILE",
        "LEU", "LYS", "MET", "PHE", "PRO", "PYL", "SER", "SEC", "THR", "TRP",
        "TYR", "VAL", "ASX", "GLX"
    ]

    n_features = len(residues)

    # FIXME: Signature of "featurize" incompatible with supertype "Featurizer"
    def featurize(  # type: ignore[override]
            self, protein_file: str,
            pockets: List[CoordinateBox]) -> np.ndarray:
        """
        Calculate atomic coodinates.

        Parameters
        ----------
        protein_file: str
            Location of PDB file. Will be loaded by MDTraj
        pockets: List[CoordinateBox]
            List of `dc.utils.CoordinateBox` objects.

        Returns
        -------
        np.ndarray
            A numpy array of shale `(len(pockets), n_residues)`
        """
        try:
            import mdtraj
        except ModuleNotFoundError:
            raise ImportError("This class requires mdtraj to be installed.")

        protein_coords = load_molecule(protein_file,
                                       add_hydrogens=False,
                                       calc_charges=False)[0]
        mapping = boxes_to_atoms(protein_coords, pockets)
        protein = mdtraj.load(protein_file)
        n_pockets = len(pockets)
        n_residues = len(BindingPocketFeaturizer.residues)
        res_map = dict(zip(BindingPocketFeaturizer.residues, range(n_residues)))
        all_features = np.zeros((n_pockets, n_residues))
        for pocket_num, pocket in enumerate(pockets):
            pocket_atoms = mapping[pocket]
            for ind, atom in enumerate(pocket_atoms):
                atom_name = str(protein.top.atom(atom))
                # atom_name is of format RESX-ATOMTYPE
                # where X is a 1 to 4 digit number
                residue = atom_name[:3]
                if residue not in res_map:
                    logger.info("Warning: Non-standard residue in PDB file")
                    continue
                all_features[pocket_num, res_map[residue]] += 1
        return all_features



"""
Gathers all datasets in one place for convenient imports
"""
# flake8: noqa
import logging

logger = logging.getLogger(__name__)

# TODO(rbharath): Get rid of * import
from deepchem.data.datasets import pad_features
from deepchem.data.datasets import pad_batch
from deepchem.data.datasets import Dataset
from deepchem.data.datasets import NumpyDataset
from deepchem.data.datasets import DiskDataset
from deepchem.data.datasets import ImageDataset
from deepchem.data.datasets import sparsify_features
from deepchem.data.datasets import densify_features
from deepchem.data.supports import *
from deepchem.data.data_loader import DataLoader
from deepchem.data.data_loader import CSVLoader
from deepchem.data.data_loader import UserCSVLoader
from deepchem.data.data_loader import JsonLoader
from deepchem.data.data_loader import SDFLoader
from deepchem.data.data_loader import FASTALoader
from deepchem.data.data_loader import FASTQLoader
from deepchem.data.data_loader import ImageLoader
from deepchem.data.data_loader import InMemoryLoader
try:
    from deepchem.data.data_loader import SAMLoader
    from deepchem.data.data_loader import BAMLoader
    from deepchem.data.data_loader import CRAMLoader
except ImportError as e:
    logger.warning(
        f'Skipped loading biological sequence loaders, missing a dependency. {e}'
    )



from __future__ import annotations
from collections import defaultdict
from typing import Optional, Tuple, List, Sequence
import re
import copy
from deepchem.utils.dft_utils.hamilton.intor.symmetry import BaseSymmetry, S1Symmetry, S4Symmetry


class IntorNameManager(object):
    """
    A class for integral name manipulation.

    This class is supposed to be string-manipulation only, so no array calculation
    should be performed here.

    Examples
    --------
    >>> mgr = IntorNameManager("int1e", "r0")
    >>> mgr.fullname
    'int1e_r0'
    >>> mgr.get_intgl_name(True)
    'int1e_r0_sph'
    >>> mgr.get_ft_intgl_name(True)
    'GTO_ft_r0_sph'
    >>> mgr.get_intgl_symmetry([0, 1, 2, 0]).code
    's1'
    >>> mgr.get_intgl_components_shape()
    (3,)
    """

    # Operators name must not contain separator name
    ops_name = ["ip", "rr"]  # Name of basis operators
    sep_name = ["a", "b"]  # Separator of basis (other than the middle operator)

    # Components shape of raw operator and basis operators
    # should be a tuple with AT MOST 1 element
    rawop_comp = defaultdict(
        tuple,  # type: ignore
        {
            "r0": (3,),
            "r0r0": (9,),
            "r0r0r0": (27,),
        })
    op_comp = defaultdict(
        tuple,  # type: ignore
        {
            "ip": (3,),
        })

    # The number of new dimensions added with the operators
    rawop_ndim = defaultdict(int, {k: len(v) for (k, v) in rawop_comp.items()})
    op_ndim = defaultdict(int, {k: len(v) for (k, v) in op_comp.items()})

    def __init__(self, int_type: str, shortname: str):
        """
        Initialize the IntorNameManager.

        Parameters
        ----------
        int_type : str
            Integral type.
        shortname : str
            Short name representing the integral.
        """
        self._int_type = int_type
        self._shortname = shortname
        self._rawop, self._ops = self.split_name(int_type, shortname)
        self._nbasis = len(self._ops)
        self._imid = (self._nbasis +
                      1) // 2  # Middle index (where the rawops should be)

    @property
    def fullname(self) -> str:
        """
        Get the full name of the integral.

        Returns
        -------
        str
            Full name of the integral.
        """
        return self._int_type + "_" + self._shortname

    @property
    def rawopname(self) -> str:
        """
        Get the raw operator name of the integral.

        Returns
        -------
        str
            Raw operator name of the integral.
        """
        return self._rawop

    @property
    def int_type(self) -> str:
        """
        Get the integral type.

        Returns
        -------
        str
            Integral type.
        """
        return self._int_type

    @property
    def shortname(self) -> str:
        """
        Get the short name representing the integral.

        Returns
        -------
        str
            Short name representing the integral.
        """
        return self._shortname

    def get_intgl_name(self, spherical: bool) -> str:
        """
        Get the full name of the integral in libcint library.

        Parameters
        ----------
        spherical : bool
            Flag indicating whether the integral is in spherical or cartesian coordinates.

        Returns
        -------
        str
            Full name of the integral in libcint library.
        """
        cartsph = "sph" if spherical else "cart"
        return f"{self.fullname}_{cartsph}"

    def get_ft_intgl_name(self, spherical: bool) -> str:
        """
        Get the full name of the fourier transform integral in libcint library.

        Parameters
        ----------
        spherical : bool
            Flag indicating whether the integral is in spherical or cartesian coordinates.

        Returns
        -------
        str
            Full name of the fourier transform integral in libcint library.
        """
        cartsph = "sph" if spherical else "cart"
        int_type = self._int_type
        if int_type == "int1e":
            return "GTO_ft_%s_%s" % (self._shortname, cartsph)
        else:
            raise NotImplementedError("Unimplemented FT integral for %s" %
                                      int_type)

    def get_intgl_deriv_namemgr(self, derivop: str,
                                ibasis: int) -> IntorNameManager:
        """
        Get the name manager of a new integral when a derivative operation is applied.

        Parameters
        ----------
        derivop : str
            String of the derivative operation.
        ibasis : int
            Index of the basis where the derivative operation should be performed (0-based).

        Returns
        -------
        IntorNameManager
            Name manager of the new integral.
        """
        assert derivop in self.ops_name
        assert ibasis < self._nbasis

        ops = copy.copy(self._ops)
        ops[ibasis] = [derivop] + ops[ibasis]
        sname = self.join_name(self._int_type, self._rawop, ops)
        return IntorNameManager(self._int_type, sname)

    def get_intgl_deriv_newaxispos(self, derivop: str,
                                   ibasis: int) -> Optional[int]:
        """
        Get the new axis position in the new integral name.

        Parameters
        ----------
        derivop : str
            Derivative operation.
        ibasis : int
            Index of the basis where the derivative operation should be performed (0-based).

        Returns
        -------
        Optional[int]
            New axis position (if any) or None if no new axis is inserted.
        """
        op_ndim = self.op_ndim[derivop]
        if op_ndim == 0:
            return None

        ops_flat: List[str] = sum(self._ops[:ibasis], [])
        new_ndim = sum([self.op_ndim[op] for op in ops_flat])

        # check if rawsname should also be included
        include_rname = ibasis >= self._imid
        if include_rname:
            new_ndim += self.rawop_ndim[self._rawop]

        return new_ndim

    def get_intgl_components_shape(self) -> Tuple[int, ...]:
        """
        Get the component shape of the array of the integral.

        Returns
        -------
        Tuple[int, ...]
            Component shape of the array of the integral.
        """
        ops_flat_l: List[str] = sum(self._ops[:self._imid], [])
        ops_flat_r: List[str] = sum(self._ops[self._imid:], [])
        comp_shape = sum([self.op_comp[op] for op in ops_flat_l], ()) + \
            self.rawop_comp[self._rawop] + \
            sum([self.op_comp[op] for op in ops_flat_r], ())
        return comp_shape

    def get_intgl_symmetry(self, unique: Sequence[int]) -> BaseSymmetry:
        """
        Get the symmetry of the integral.

        Parameters
        ----------
        unique : Sequence[int]
            Sequence representing the uniqueness of each dimension.

        Returns
        -------
        BaseSymmetry
            Symmetry of the integral.
        """
        all_same = all([u == 0 for u in unique])
        if all_same:
            if self._int_type == "int2e":
                if self._shortname == "ar12b":
                    return S4Symmetry()
        return S1Symmetry()

    def get_transpose_path_to(
            self, other: IntorNameManager) -> Optional[List[Tuple[int, int]]]:
        """
        Check if the integration `other` can be achieved by transposing `self`.

        Parameters
        ----------
        other : IntorNameManager
            IntorNameManager instance representing the other integral.

        Returns
        -------
        Optional[List[Tuple[int, int]]]
            List of two dims if it can for the transpose-path of `self`
            to get the same result as `other`, None if it cannot.
        """
        nbasis = self._nbasis
        # get the basis transpose paths
        if nbasis == 2:
            transpose_paths: List[List[Tuple[int, int]]] = [
                [],
                [(-1, -2)],
            ]
        elif nbasis == 3:
            # note: the third basis is usually an auxiliary basis which typically
            # different from the first two
            transpose_paths = [
                [],
                [(-2, -3)],
            ]
        elif nbasis == 4:
            transpose_paths = [
                [],
                [(-3, -4)],
                [(-1, -2)],
                [(-1, -3), (-2, -4)],
                [(-1, -3), (-2, -4), (-2, -1)],
                [(-1, -3), (-2, -4), (-3, -4)],
            ]
        else:
            raise self._nbasis_error(nbasis)

        def _swap(p: List[List[str]],
                  path: List[Tuple[int, int]]) -> List[List[str]]:
            # swap the pattern according to the given transpose path
            r = p[:]  # make a copy
            for i0, i1 in path:
                r[i0], r[i1] = r[i1], r[i0]
            return r

        # try all the transpose path until gets a match
        for transpose_path in transpose_paths:
            if _swap(self._ops, transpose_path) == other._ops:
                return transpose_path
        return None

    def get_comp_permute_path(
            self, transpose_path: List[Tuple[int, int]]) -> List[int]:
        """
        Get the component permute path given the basis transpose path.

        Parameters
        ----------
        transpose_path : List[Tuple[int, int]]
            Basis transpose path.

        Returns
        -------
        List[int]
            Component permute path.
        """
        flat_ops: List[str] = sum(self._ops, [])
        _ = flat_ops.count("ip")  # number of ip operators

        # get the positions of the axes
        dim_pos = []
        ioffset = 0
        for i, ops in enumerate(self._ops):
            if i == self._imid:
                naxes = self.rawop_ndim[self._rawop]
                dim_pos.append(list(range(ioffset, ioffset + naxes)))
                ioffset += naxes
            naxes = sum([self.op_ndim[op] for op in ops])
            dim_pos.append(list(range(ioffset, ioffset + naxes)))
            ioffset += naxes

        # add the bases' axes (assuming each basis only occupy one axes)
        for i in range(self._nbasis):
            dim_pos.append([ioffset])
            ioffset += 1

        # swap the axes
        for t0, t1 in transpose_path:
            s0 = t0 + self._nbasis
            s1 = t1 + self._nbasis
            s0 += 1 if s0 >= self._imid else 0
            s1 += 1 if s1 >= self._imid else 0
            dim_pos[s0], dim_pos[s1] = dim_pos[s1], dim_pos[s0]

        # flatten the list to get the permutation path
        dim_pos_flat: List[int] = sum(dim_pos, [])
        return dim_pos_flat

    @classmethod
    def split_name(cls, int_type: str,
                   shortname: str) -> Tuple[str, List[List[str]]]:
        """
        Split the shortname into operator per basis and return the raw shortname as well.

        Parameters
        ----------
        int_type : str
            Integral type.
        shortname : str
            Short name representing the integral.

        Returns
        -------
        Tuple[str, List[List[str]]]
            Raw shortname and list of basis operators.
        """
        deriv_ops = cls.ops_name
        deriv_pattern = re.compile("(" + ("|".join(deriv_ops)) + ")")

        # get the raw shortname (i.e. shortname without derivative operators)
        rawsname = shortname
        for op in deriv_ops:
            rawsname = rawsname.replace(op, "")

        nbasis = cls.get_nbasis(int_type)
        if nbasis == 2:
            ops_str = shortname.split(rawsname)
        elif nbasis == 3:
            assert rawsname.startswith("a"), rawsname
            rawsname = rawsname[1:]
            ops_l, ops_r = shortname.split(rawsname)
            ops_l1, ops_l2 = ops_l.split("a")
            ops_str = [ops_l1, ops_l2, ops_r]
        elif nbasis == 4:
            assert rawsname.startswith("a") and rawsname.endswith("b"), rawsname
            rawsname = rawsname[1:-1]
            ops_l, ops_r = shortname.split(rawsname)
            ops_l1, ops_l2 = ops_l.split("a")
            ops_r1, ops_r2 = ops_r.split("b")
            ops_str = [ops_l1, ops_l2, ops_r1, ops_r2]
        else:
            raise cls._nbasis_error(nbasis)

        ops = [re.findall(deriv_pattern, op_str) for op_str in ops_str]
        assert len(ops) == nbasis
        return rawsname, ops

    @classmethod
    def join_name(cls, int_type: str, rawsname: str,
                  ops: List[List[str]]) -> str:
        """
        Get the shortname given rawsname and list of basis ops.

        Parameters
        ----------
        int_type : str
            Integral type.
        rawsname : str
            Raw shortname.
        ops : List[List[str]]
            List of basis operators.

        Returns
        -------
        str
            Shortname representing the integral.
        """
        nbasis = cls.get_nbasis(int_type)
        ops_str = ["".join(op) for op in ops]
        assert len(ops_str) == nbasis

        if nbasis == 2:
            return ops_str[0] + rawsname + ops_str[1]
        elif nbasis == 3:
            return ops_str[0] + cls.sep_name[0] + ops_str[
                1] + rawsname + ops_str[2]
        elif nbasis == 4:
            return ops_str[0] + cls.sep_name[0] + ops_str[1] + rawsname + ops_str[2] + \
                cls.sep_name[1] + ops_str[3]
        else:
            raise cls._nbasis_error(nbasis)

    @classmethod
    def get_nbasis(cls, int_type: str) -> int:
        """
        Get the number of basis according to the integral type.

        Parameters
        ----------
        int_type : str
            Integral type.

        Returns
        -------
        int
            Number of basis.
        """
        if int_type == "int1e" or int_type == "int2c2e":
            return 2
        elif int_type == "int3c2e":
            return 3
        elif int_type == "int2e":
            return 4
        else:
            raise RuntimeError(f"Unknown integral type: {int_type}")

    @classmethod
    def _nbasis_error(cls, nbasis: int):
        return RuntimeError(f"Unknown integral with {nbasis} basis")



import numpy as np
from typing import Optional
from deepchem.feat import Featurizer


class SAMFeaturizer(Featurizer):
    """
    Featurizes SAM files, that store biological sequences aligned to a reference
    sequence. This class extracts Query Name, Query Sequence, Query Length,
    Reference Name,Reference Start, CIGAR and Mapping Quality of each read in
    a SAM file.

    This is the default featurizer used by SAMLoader, and it extracts the following
    fields from each read in each SAM file in the given order:-
    - Column 0: Query Name
    - Column 1: Query Sequence
    - Column 2: Query Length
    - Column 3: Reference Name
    - Column 4: Reference Start
    - Column 5: CIGAR
    - Column 6: Mapping Quality

    Examples
    --------
    >>> from deepchem.data.data_loader import SAMLoader
    >>> import deepchem as dc
    >>> inputs = 'deepchem/data/tests/example.sam'
    >>> featurizer = dc.feat.SAMFeaturizer()
    >>> features = featurizer.featurize(inputs)
    >>> type(features[0])
    <class 'numpy.ndarray'>

    Note
    ----
    This class requires pysam to be installed. Pysam can be used with Linux or MacOS X.
    To use Pysam on Windows, use Windows Subsystem for Linux(WSL).

    """

    def __init__(self, max_records=None):
        """
        Initialize SAMFeaturizer.

        Parameters
        ----------
        max_records : int or None, optional
            The maximum number of records to extract from the SAM file. If None, all records will be extracted.

        """
        self.max_records = max_records

    def _featurize(self, datapoint):
        """
        Extract features from a SAM file.

        Parameters
        ----------
        datapoint : str
            Name of SAM file.

        Returns
        -------
        features : numpy.ndarray
        A 2D NumPy array representing the extracted features.
        Each row corresponds to a SAM record, and columns represent different features.
            - Column 0: Query Name
            - Column 1: Query Sequence
            - Column 2: Query Length
            - Column 3: Reference Name
            - Column 4: Reference Start
            - Column 5: CIGAR
            - Column 6: Mapping Quality

        """

        features = []
        record_count = 0

        for record in datapoint:
            feature_vector = [
                record.query_name,
                record.query_sequence,
                record.query_length,
                record.reference_name,
                record.reference_start,
                record.cigar,
                record.mapping_quality,
            ]

            features.append(feature_vector)
            record_count += 1

            # Break the loop if max_records is set
            if self.max_records is not None and record_count >= self.max_records:
                break

        datapoint.close()

        return np.array(features, dtype="object")


class BAMFeaturizer(Featurizer):
    """
    Featurizes BAM files, that are compressed binary representations of SAM
    (Sequence Alignment Map) files. This class extracts Query Name, Query
    Sequence, Query Length, Reference Name, Reference Start, CIGAR and Mapping
    Quality of the alignment in the BAM file.

    This is the default featurizer used by BAMLoader, and it extracts the
    following fields from each read in each BAM file in the given order:-
    - Column 0: Query Name
    - Column 1: Query Sequence
    - Column 2: Query Length
    - Column 3: Reference Name
    - Column 4: Reference Start
    - Column 5: CIGAR
    - Column 6: Mapping Quality
    - Column 7: Pileup Information (if get_pileup=True)

    Additionally, we can also get pileups from BAM files by setting
    get_pileup=True.A pileup is a summary of the alignment of reads
    at each position in a reference sequence. Specifically, it
    provides information on the position on the reference genome,
    the depth of coverage (i.e., the number of reads aligned to that
    position), and the actual bases from the aligned reads at that
    position, along with their quality scores. This data structure
    is useful for identifying variations, such as single nucleotide
    polymorphisms (SNPs), insertions, and deletions by comparing the
    aligned reads to the reference genome. A pileup can be visualized
    as a vertical stack of aligned sequences, showing how each read
    matches or mismatches the reference at each position.
    In DeepVariant, pileups are utilized during the initial stages to
    select candidate windows for further analysis.

    Examples
    --------
    >>> from deepchem.data.data_loader import BAMLoader
    >>> import deepchem as dc
    >>> inputs = 'deepchem/data/tests/example.bam'
    >>> featurizer = dc.feat.BAMFeaturizer()
    >>> features = featurizer.featurize(inputs)
    >>> type(features[0])
    <class 'numpy.ndarray'>

    Note
    ----
    This class requires pysam to be installed. Pysam can be used with Linux or MacOS X.
    To use Pysam on Windows, use Windows Subsystem for Linux(WSL).

    """

    def __init__(self, max_records=None, get_pileup: Optional[bool] = False):
        """
        Initialize BAMFeaturizer.

        Parameters
        ----------
        max_records : int or None, optional
            The maximum number of records to extract from the BAM file. If None, all
            records will be extracted.
        get_pileup : bool, optional
            If True, pileup information will be extracted from the BAM file.
            This is used in DeepVariant. False by default.

        """
        self.max_records = max_records
        self.get_pileup = get_pileup

    def _featurize(self, datapoint):
        """
        Extract features from a BAM file.

        Parameters
        ----------
        datapoint : str
            Name of the BAM file.
            The corresponding index file must be in the same directory.

        Returns
        -------
        features : numpy.ndarray
        A 2D NumPy array representing the extracted features.

        """

        features = []
        record_count = 0

        for record in datapoint:
            initial_position = datapoint.tell()
            feature_vector = [
                record.query_name,
                record.query_sequence,
                record.query_length,
                record.reference_name,
                record.reference_start,
                record.cigar,
                record.mapping_quality,
            ]

            if (self.get_pileup):
                pileup_columns = []
                for pileupcolumn in datapoint.pileup():
                    pileup_info = {
                        "name":
                            pileupcolumn.reference_name,
                        "pos":
                            pileupcolumn.reference_pos,
                        "depth":
                            pileupcolumn.nsegments,
                        "reads": [[
                            pileupread.alignment.query_sequence,
                            pileupread.query_position, pileupread.is_del,
                            pileupread.is_refskip, pileupread.indel
                        ] for pileupread in pileupcolumn.pileups]
                    }
                    pileup_columns.append(pileup_info)
                feature_vector.append(pileup_columns)

            features.append(feature_vector)
            record_count += 1
            datapoint.seek(initial_position)

            # Break the loop if max_records is set
            if self.max_records is not None and record_count >= self.max_records:
                break

        datapoint.close()

        return np.array(features, dtype="object")


class CRAMFeaturizer(Featurizer):
    """
    Featurizes CRAM files, that are compressed columnar file format for storing
    biological sequences aligned to a reference sequence. This class extracts Query Name, Query
    Sequence, Query Length, Reference Name, Reference Start, CIGAR and Mapping
    Quality of the alignment in the CRAM file.

    This is the default featurizer used by CRAMLoader, and it extracts the following
    fields from each read in each CRAM file in the given order:-
    - Column 0: Query Name
    - Column 1: Query Sequence
    - Column 2: Query Length
    - Column 3: Reference Name
    - Column 4: Reference Start
    - Column 5: CIGAR
    - Column 6: Mapping Quality

    Examples
    --------
    >>> from deepchem.data.data_loader import CRAMLoader
    >>> import deepchem as dc
    >>> inputs = 'deepchem/data/tests/example.cram'
    >>> featurizer = dc.feat.CRAMFeaturizer()
    >>> features = featurizer.featurize(inputs)
    >>> type(features[0])
    <class 'numpy.ndarray'>

    Note
    ----
    This class requires pysam to be installed. Pysam can be used with Linux or MacOS X.
    To use Pysam on Windows, use Windows Subsystem for Linux(WSL).

    """

    def __init__(self, max_records=None):
        """
        Initialize CRAMFeaturizer.

        Parameters
        ----------
        max_records : int or None, optional
            The maximum number of records to extract from the CRAM file. If None, all
            records will be extracted.

        """
        self.max_records = max_records

    def _featurize(self, datapoint):
        """
        Extract features from a CRAM file.

        Parameters
        ----------
        datapoint : str
            Name of the CRAM file.

        Returns
        -------
        features : numpy.ndarray
        A 2D NumPy array representing the extracted features.

        """

        features = []
        record_count = 0

        for record in datapoint:
            feature_vector = [
                record.query_name,
                record.query_sequence,
                record.query_length,
                record.reference_name,
                record.reference_start,
                record.cigar,
                record.mapping_quality,
            ]

            features.append(feature_vector)
            record_count += 1

            # Break the loop if max_records is set
            if self.max_records is not None and record_count >= self.max_records:
                break

        datapoint.close()

        return np.array(features, dtype="object")



from typing import List, Callable, Sequence, NamedTuple, Union
import torch


# list of tableaus
class _Tableau(NamedTuple):
    """To specify a particular method, one needs to provide the integer s
    (the number of stages), and the coefficients a[i,j] (for 1 ≤ j < i ≤ s),
    b[i] (for i = 1, 2, ..., s) and c[i] (for i = 2, 3, ..., s). The matrix
    [aij] is called the Runge–Kutta matrix, while the b[i] and c[i] are known
    as the weights and the nodes. These data are usually arranged in a
    mnemonic device, known as a Butcher tableau (after John C. Butcher):

    Examples
    --------
    >>> euler = _Tableau(c=[0.0],
    ...                  b=[1.0],
    ...                  a=[[0.0]]
    ... )
    >>> euler.c
    [0.0]

    Attributes
    ----------
    c: List[float]
        The nodes
    b: List[float]
        The weights
    a: List[List[float]]
        The Runge-Kutta matrix

    """
    c: List[float]
    b: List[float]
    a: List[List[float]]


rk4_tableau = _Tableau(c=[0.0, 0.5, 0.5, 1.0],
                       b=[1 / 6., 1 / 3., 1 / 3., 1 / 6.],
                       a=[[0.0, 0.0, 0.0, 0.0], [0.5, 0.0, 0.0, 0.0],
                          [0.0, 0.5, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]])
rk38_tableau = _Tableau(c=[0.0, 1 / 3, 2 / 3, 1.0],
                        b=[1 / 8, 3 / 8, 3 / 8, 1 / 8],
                        a=[[0.0, 0.0, 0.0, 0.0], [1 / 3, 0.0, 0.0, 0.0],
                           [-1 / 3, 1.0, 0.0, 0.0], [1.0, -1.0, 1.0, 0.0]])
fwd_euler_tableau = _Tableau(c=[0.0], b=[1.0], a=[[0.0]])
midpoint_fableau = _Tableau(c=[0.0, 0.5],
                            b=[0.0, 1.0],
                            a=[[0.0, 0.0], [0.5, 0.0]])


def explicit_rk(tableau: _Tableau, fcn: Callable[..., torch.Tensor],
                y0: torch.Tensor, t: torch.Tensor,
                params: Sequence[torch.Tensor]):
    """The family of explicit Runge–Kutta methods is a generalization
    of the RK4 method mentioned above.

    Examples
    --------
    >>> def lotka_volterra(y, x, params):
    ...     y1, y2 = y
    ...     a, b, c, d = params
    ...     return torch.tensor([(a * y1 - b * y1 * y2), (c * y2 * y1 - d * y2)])
    >>> t = torch.linspace(0, 50, 100)
    >>> solver_param = [rk4_tableau,
    ...                 lotka_volterra,
    ...                 torch.tensor([10., 1.]),
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> sol = explicit_rk(*solver_param)
    >>> sol[-1]
    tensor([0.3459, 3.2954])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    References
    ----------
    [1].. https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods#Explicit_Runge.E2.80.93Kutta_methods

    """
    c = tableau.c
    a = tableau.a
    b = tableau.b
    s = len(c)
    nt = len(t)

    # set up the results list
    yt_lst: List[torch.Tensor] = []
    yt_lst.append(y0)
    y = yt_lst[-1]
    for i in range(nt - 1):
        t0 = t[i]
        t1 = t[i + 1]
        h = t1 - t0
        ks: List[torch.Tensor] = []
        ksum: Union[float, torch.Tensor] = 0.0
        for j in range(s):
            if j == 0:
                k = fcn(y, t0, params)
            else:
                ak: Union[float, torch.Tensor] = 0.0
                aj = a[j]
                for m in range(j):
                    ak = aj[m] * ks[m] + ak
                k = fcn(h * ak + y, t0 + c[j] * h, params)
            ks.append(k)
            ksum = ksum + b[j] * k
        y = h * ksum + y
        yt_lst.append(y)
    yt = torch.stack(yt_lst, dim=0)
    return yt


# list of methods
def rk38_ivp(fcn: Callable[..., torch.Tensor], y0: torch.Tensor,
             t: torch.Tensor, params: Sequence[torch.Tensor], **kwargs):
    """A slight variation of "the" Runge–Kutta method is also due to
    Kutta in 1901 and is called the 3/8-rule.[19] The primary advantage
    this method has is that almost all of the error coefficients are
    smaller than in the popular method, but it requires slightly more
    FLOPs (floating-point operations) per time step.

    Examples
    --------
    >>> def lotka_volterra(y, x, params):
    ...     y1, y2 = y
    ...     a, b, c, d = params
    ...     return torch.tensor([(a * y1 - b * y1 * y2), (c * y2 * y1 - d * y2)])
    >>> t = torch.linspace(0, 50, 100)
    >>> solver_param = [lotka_volterra,
    ...                 torch.tensor([10., 1.]),
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> sol = rk38_ivp(*solver_param)
    >>> sol[-1]
    tensor([0.3483, 3.2585])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(rk38_tableau, fcn, y0, t, params)


def fwd_euler_ivp(fcn: Callable[..., torch.Tensor], y0: torch.Tensor,
                  t: torch.Tensor, params: Sequence[torch.Tensor], **kwargs):
    """However, the simplest Runge–Kutta method is the (forward) Euler method,
    given by the formula $y_{n+1} = y_{n} + hf(t_{n}, y_{n}). This is the only
    consistent explicit Runge–Kutta method with one stage.

    Examples
    --------
    >>> def lotka_volterra(y, x, params):
    ...     y1, y2 = y
    ...     a, b, c, d = params
    ...     return torch.tensor([(a * y1 - b * y1 * y2), (c * y2 * y1 - d * y2)])
    >>> t = torch.linspace(0, 50, 1000)
    >>> solver_param = [lotka_volterra,
    ...                 torch.tensor([10., 1.]),
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> sol = fwd_euler_ivp(*solver_param)
    >>> sol[-1]
    tensor([4.7419, 0.2445])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(fwd_euler_tableau, fcn, y0, t, params)


def rk4_ivp(fcn: Callable[..., torch.Tensor], y0: torch.Tensor, t: torch.Tensor,
            params: Sequence[torch.Tensor], **kwargs):
    """The most commonly used Runge Kutta method to find the solution
    of a differential equation is the RK4 method, i.e., the fourth-order
    Runge-Kutta method. The Runge-Kutta method provides the approximate
    value of y for a given point x. Only the first order ODEs can be
    solved using the Runge Kutta RK4 method.

    Examples
    --------
    >>> def lotka_volterra(y, x, params):
    ...     y1, y2 = y
    ...     a, b, c, d = params
    ...     return torch.tensor([(a * y1 - b * y1 * y2), (c * y2 * y1 - d * y2)])
    >>> t = torch.linspace(0, 50, 100)
    >>> solver_param = [lotka_volterra,
    ...                 torch.tensor([10., 1.]),
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> sol = rk4_ivp(*solver_param)
    >>> sol[-1]
    tensor([0.3459, 3.2954])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(rk4_tableau, fcn, y0, t, params)


def mid_point_ivp(fcn: Callable[..., torch.Tensor], y0: torch.Tensor,
                  t: torch.Tensor, params: Sequence[torch.Tensor], **kwargs):
    """The explicit midpoint method is sometimes also known as the
    modified Euler method, the implicit method is the most simple
    collocation method, and, applied to Hamiltonian dynamics, a
    symplectic integrator.

    Examples
    --------
    >>> def lotka_volterra(y, x, params):
    ...     y1, y2 = y
    ...     a, b, c, d = params
    ...     return torch.tensor([(a * y1 - b * y1 * y2), (c * y2 * y1 - d * y2)])
    >>> t = torch.linspace(0, 50, 100)
    >>> solver_param = [lotka_volterra,
    ...                 torch.tensor([10., 1.]),
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> sol = rk4_ivp(*solver_param)
    >>> sol[-1]
    tensor([0.3459, 3.2954])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(midpoint_fableau, fcn, y0, t, params)



"""
KAGGLE dataset loader.
"""
import os
import logging
import time

import deepchem
from deepchem.molnet.load_function.kaggle_features import merck_descriptors
from deepchem.utils import remove_missing_entries

logger = logging.getLogger(__name__)


def get_transformers(train_dataset):
    """Get transformers applied to datasets."""
    transformers = []
    # transformers = [
    #    deepchem.trans.LogTransformer(transform_X=True),
    #    deepchem.trans.NormalizationTransformer(transform_y=True,
    #                                      dataset=train_dataset)]
    return transformers


# Set shard size low to avoid memory problems.
def gen_kaggle(KAGGLE_tasks,
               train_dir,
               valid_dir,
               test_dir,
               data_dir,
               shard_size=2000):
    """Load KAGGLE datasets. Does not do train/test split"""
    # TIMING
    time1 = time.time()
    # TIMING
    # Set some global variables up top
    train_files = os.path.join(
        data_dir, "KAGGLE_training_disguised_combined_full.csv.gz")
    valid_files = os.path.join(data_dir,
                               "KAGGLE_test1_disguised_combined_full.csv.gz")
    test_files = os.path.join(data_dir,
                              "KAGGLE_test2_disguised_combined_full.csv.gz")
    if not os.path.exists(train_files):
        deepchem.utils.data_utils.download_url(
            "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/KAGGLE_training_disguised_combined_full.csv.gz",
            dest_dir=data_dir)
        deepchem.utils.data_utils.download_url(
            "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/KAGGLE_test1_disguised_combined_full.csv.gz",
            dest_dir=data_dir)
        deepchem.utils.data_utils.download_url(
            "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/KAGGLE_test2_disguised_combined_full.csv.gz",
            dest_dir=data_dir)

    # Featurize KAGGLE dataset
    logger.info("About to featurize KAGGLE dataset.")
    featurizer = deepchem.feat.UserDefinedFeaturizer(merck_descriptors)

    loader = deepchem.data.UserCSVLoader(tasks=KAGGLE_tasks,
                                         id_field="Molecule",
                                         featurizer=featurizer)

    logger.info("Featurizing train datasets")
    train_dataset = loader.featurize(train_files, shard_size=shard_size)

    logger.info("Featurizing valid datasets")
    valid_dataset = loader.featurize(valid_files, shard_size=shard_size)

    logger.info("Featurizing test datasets")
    test_dataset = loader.featurize(test_files, shard_size=shard_size)

    logger.info("Remove missing entries from datasets.")
    remove_missing_entries(train_dataset)
    remove_missing_entries(valid_dataset)
    remove_missing_entries(test_dataset)

    logger.info("Shuffling order of train dataset.")
    train_dataset.sparse_shuffle()

    logger.info("Transforming datasets with transformers.")
    transformers = get_transformers(train_dataset)

    for transformer in transformers:
        logger.info("Performing transformations with %s" %
                    transformer.__class__.__name__)
        logger.info("Transforming datasets")
        train_dataset = transformer.transform(train_dataset)
        valid_dataset = transformer.transform(valid_dataset)
        test_dataset = transformer.transform(test_dataset)

    logger.info("Moving directories")
    train_dataset.move(train_dir)
    valid_dataset.move(valid_dir)
    test_dataset.move(test_dir)

    # TIMING
    time2 = time.time()
    logger.info("TIMING: KAGGLE fitting took %0.3f s" % (time2 - time1))
    # TIMING

    return train_dataset, valid_dataset, test_dataset


def load_kaggle(shard_size=2000, featurizer=None, split=None, reload=True):
    """Loads kaggle datasets. Generates if not stored already.

    The Kaggle dataset is an in-house dataset from Merck that was first introduced in the following paper:

    Ma, Junshui, et al. "Deep neural nets as a method for quantitative structure–activity relationships." Journal of chemical information and modeling 55.2 (2015): 263-274.

    It contains 100,000 unique Merck in-house compounds that were
    measured on 15 enzyme inhibition and ADME/TOX datasets.
    Unlike most of the other datasets featured in MoleculeNet,
    the Kaggle collection does not have structures for the
    compounds tested since they were proprietary Merck compounds.
    However, the collection does feature pre-computed descriptors
    for these compounds.

    Note that the original train/valid/test split from the source
    data was preserved here, so this function doesn't allow for
    alternate modes of splitting. Similarly, since the source data
    came pre-featurized, it is not possible to apply alternative
    featurizations.

    Parameters
    ----------
    shard_size: int, optional
        Size of the DiskDataset shards to write on disk
    featurizer: optional
        Ignored since featurization pre-computed
    split: optional
        Ignored since split pre-computed
    reload: bool, optional
        Whether to automatically re-load from disk

    """
    KAGGLE_tasks = [
        '3A4', 'CB1', 'DPP4', 'HIVINT', 'HIV_PROT', 'LOGD', 'METAB', 'NK1',
        'OX1', 'OX2', 'PGP', 'PPB', 'RAT_F', 'TDI', 'THROMBIN'
    ]
    data_dir = deepchem.utils.data_utils.get_data_dir()

    data_dir = os.path.join(data_dir, "kaggle")
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)
    train_dir = os.path.join(data_dir, "train_dir")
    valid_dir = os.path.join(data_dir, "valid_dir")
    test_dir = os.path.join(data_dir, "test_dir")

    if (os.path.exists(train_dir) and os.path.exists(valid_dir) and
            os.path.exists(test_dir)):
        logger.info("Reloading existing datasets")
        train_dataset = deepchem.data.DiskDataset(train_dir)
        valid_dataset = deepchem.data.DiskDataset(valid_dir)
        test_dataset = deepchem.data.DiskDataset(test_dir)
    else:
        logger.info("Featurizing datasets")
        train_dataset, valid_dataset, test_dataset = \
          gen_kaggle(KAGGLE_tasks, train_dir, valid_dir, test_dir, data_dir,
                      shard_size=shard_size)

    transformers = get_transformers(train_dataset)
    return KAGGLE_tasks, (train_dataset, valid_dataset,
                          test_dataset), transformers
